#
# Copyright (C) 2018  FreeIPA Contributors see COPYING for license
#
"""Tests for ipalib.util module
"""

import os
import ssl
from unittest import mock

import pytest

from ipalib.util import (
    get_pager, create_https_connection, get_proper_tls_version_span
)
from ipaplatform.constants import constants


@pytest.mark.parametrize('pager,expected_result', [
    # Valid values
    ('cat', '/bin/cat'),
    ('/bin/cat', '/bin/cat'),
    # Invalid values (wrong command, package is not installed, etc)
    ('cat_', None),
    ('', None)
])
def test_get_pager(pager, expected_result):
    with mock.patch.dict(os.environ, {'PAGER': pager}):
        pager = get_pager()
        assert(pager == expected_result or pager.endswith(expected_result))


BASE_CTX = ssl.SSLContext(ssl.PROTOCOL_TLS)
if constants.TLS_HIGH_CIPHERS is not None:
    BASE_CTX.set_ciphers(constants.TLS_HIGH_CIPHERS)
else:
    BASE_CTX.set_ciphers("PROFILE=SYSTEM")

# options: IPA still supports Python 3.6 without min/max version setters
BASE_OPT = BASE_CTX.options
BASE_OPT |= (
    ssl.OP_ALL | ssl.OP_NO_COMPRESSION | ssl.OP_SINGLE_DH_USE |
    ssl.OP_SINGLE_ECDH_USE
)
TLS_OPT = (
    ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 | ssl.OP_NO_TLSv1 |
    ssl.OP_NO_TLSv1_1
)
OP_NO_TLSv1_3 = getattr(ssl, "OP_NO_TLSv1_3", 0)  # make pylint happy


@pytest.mark.skip_if_platform(
    "debian", reason="Crypto policy is not supported on Debian"
)
@pytest.mark.parametrize('minver,maxver,opt,expected', [
    (None, None, BASE_OPT, None),
    (None, "tls1.3", BASE_OPT | TLS_OPT, ["tls1.2", "tls1.3"]),
    ("tls1.2", "tls1.3", BASE_OPT | TLS_OPT, ["tls1.2", "tls1.3"]),
    ("tls1.2", None, BASE_OPT | TLS_OPT, ["tls1.2", "tls1.3"]),
    ("tls1.2", "tls1.2", BASE_OPT | TLS_OPT | OP_NO_TLSv1_3, ["tls1.2"]),
    (None, "tls1.2", BASE_OPT | TLS_OPT | OP_NO_TLSv1_3, ["tls1.2"]),
    ("tls1.3", "tls1.3", BASE_OPT | TLS_OPT | ssl.OP_NO_TLSv1_2, ["tls1.3"]),
    ("tls1.3", None, BASE_OPT | TLS_OPT | ssl.OP_NO_TLSv1_2, ["tls1.3"]),
])
def test_tls_version_span(minver, maxver, opt, expected):
    assert get_proper_tls_version_span(minver, maxver) == expected
    # file must exist and contain certs
    cafile = ssl.get_default_verify_paths().cafile
    conn = create_https_connection(
        "invalid.test",
        cafile=cafile,
        tls_version_min=minver,
        tls_version_max=maxver
    )
    ctx = getattr(conn, "_context")
    assert ctx.options == BASE_OPT | opt
    assert ctx.get_ciphers() == BASE_CTX.get_ciphers()



#
# Copyright (C) 2020  FreeIPA Contributors see COPYING for license
#

import time

from cryptography.hazmat.backends import default_backend
from cryptography import x509
import pytest

from ipalib.constants import IPA_CA_RECORD
from ipatests.test_integration.base import IntegrationTest
from ipatests.pytest_ipa.integration import tasks
from ipatests.test_integration.test_caless import CALessBase, ipa_certs_cleanup
from ipaplatform.osinfo import osinfo
from ipaplatform.paths import paths
from ipatests.test_integration.test_external_ca import (
    install_server_external_ca_step1,
    install_server_external_ca_step2,
)


IPA_CA = "ipa_ca.crt"
ROOT_CA = "root_ca.crt"

# RHEL does not have certbot.  EPEL's version is broken with
# python-cryptography-2.3; likewise recent PyPI releases.
# So for now, on RHEL we suppress tests that use certbot.
skip_certbot_tests = osinfo.id not in ['fedora', 'rhel']

# Fedora mod_md package needs some patches before it will work.
# RHEL version has the patches.
skip_mod_md_tests = osinfo.id not in ['rhel', 'fedora', ]

CERTBOT_DNS_IPA_SCRIPT = '/usr/libexec/ipa/acme/certbot-dns-ipa'


def check_acme_status(host, exp_status, timeout=60):
    """Helper method to check the status of acme server"""
    for _i in range(0, timeout, 5):
        result = host.run_command(['ipa-acme-manage', 'status'])
        status = result.stdout_text.split(" ")[2].strip()
        print("ACME status: %s" % status)
        if status == exp_status:
            break
        time.sleep(5)
    else:
        raise RuntimeError("request timed out")

    return status


def get_selinux_status(host):
    """
    Return the SELinux enforcing status.

    Return True if enabled and enforcing, otherwise False
    """
    result = host.run_command(['/usr/sbin/selinuxenabled'], raiseonerr=False)
    if result.returncode != 0:
        return False

    result = host.run_command(['/usr/sbin/getenforce'], raiseonerr=False)
    if 'Enforcing' in result.stdout_text:
        return True

    return False


def server_install_teardown(func):
    def wrapped(*args):
        master = args[0].master
        try:
            func(*args)
        finally:
            ipa_certs_cleanup(master)
    return wrapped


def prepare_acme_client(master, client):
    # cache the acme service uri
    acme_host = f'{IPA_CA_RECORD}.{master.domain.name}'
    acme_server = f'https://{acme_host}/acme/directory'

    # install acme client packages
    if not skip_certbot_tests:
        tasks.install_packages(client, ['certbot'])
    if not skip_mod_md_tests:
        tasks.install_packages(client, ['mod_md'])

    return acme_server


def certbot_register(host, acme_server):
    """method to register the host to acme server"""
    # clean up any existing registration and certificates
    host.run_command(
        [
            'rm', '-rf',
            '/etc/letsencrypt/accounts',
            '/etc/letsencrypt/archive',
            '/etc/letsencrypt/csr',
            '/etc/letsencrypt/keys',
            '/etc/letsencrypt/live',
            '/etc/letsencrypt/renewal',
            '/etc/letsencrypt/renewal-hooks'
        ]
    )
    # service is enabled; registration should succeed
    host.run_command(
        [
            'certbot',
            '--server', acme_server,
            'register',
            '-m', 'nobody@example.test',
            '--agree-tos',
            '--no-eff-email',
        ],
    )


def certbot_standalone_cert(host, acme_server):
    """method to issue a certbot's certonly standalone cert"""
    # Get a cert from ACME service using HTTP challenge and Certbot's
    # standalone HTTP server mode
    host.run_command(['systemctl', 'stop', 'httpd'])
    host.run_command(
        [
            'certbot',
            '--server', acme_server,
            'certonly',
            '--domain', host.hostname,
            '--standalone',
        ]
    )


class TestACME(CALessBase):
    """
    Test the FreeIPA ACME service by using ACME clients on a FreeIPA client.

    We currently test:

        * service enable/disable (using Curl)
        * http-01 challenge with Certbot's standalone HTTP server
        * dns-01 challenge with Certbot and FreeIPA DNS via hook scripts
        * revocation with Certbot
        * http-01 challenge with mod_md

    Tests we should add:

        * dns-01 challenge with mod_md (see
          https://httpd.apache.org/docs/current/mod/mod_md.html#mdchallengedns01)

    Things that are not implmented/supported yet, but may be in future:

        * IP address SAN
        * tls-alpn-01 challenge
        * Other clients or service scenarios

    """
    num_replicas = 1
    num_clients = 1

    @classmethod
    def install(cls, mh):
        super(TestACME, cls).install(mh)

        # install packages before client install in case of IPA DNS problems
        cls.acme_server = prepare_acme_client(cls.master, cls.clients[0])

        # Each subclass handles its own server installation procedure
        if cls.__name__ != 'TestACME':
            return

        tasks.install_master(cls.master, setup_dns=True)

        tasks.install_client(cls.master, cls.clients[0])
        tasks.install_replica(cls.master, cls.replicas[0])

    def certinstall(self, certfile=None, keyfile=None,
                    pin=None):
        """Small wrapper around ipa-server-certinstall

           We are always replacing only the web server with a fixed
           pre-generated value and returning the result for the caller
           to figure out.
        """
        self.create_pkcs12('ca1/server', password=None, filename='server.p12')
        self.copy_cert(self.master, 'server.p12')
        if pin is None:
            pin = self.cert_password
        args = ['ipa-server-certinstall',
                '-p', self.master.config.dirman_password,
                '--pin', pin,
                '-w']
        if certfile:  # implies keyfile
            args.append(certfile)
            args.append(keyfile)
        else:
            args.append('server.p12')

        return self.master.run_command(args,
                                       raiseonerr=False)

    #######
    # kinit
    #######

    def test_kinit_master(self):
        # Some tests require executing ipa commands, e.g. to
        # check revocation status or add/remove DNS entries.
        # Preemptively kinit as admin on the master.
        tasks.kinit_admin(self.master)

    #####################
    # Enable ACME service
    #####################

    def test_acme_service_not_yet_enabled(self):
        # --fail makes curl exit code 22 when response status >= 400.
        # ACME service should return 503 because it was not enabled yet.
        self.clients[0].run_command(
            ['curl', '--fail', self.acme_server],
            ok_returncode=22,
        )
        result = self.master.run_command(['ipa-acme-manage', 'status'])
        assert 'disabled' in result.stdout_text

    def test_enable_acme_service(self):
        self.master.run_command(['ipa-acme-manage', 'enable'])

        # wait a short time for Dogtag ACME service to observe config
        # change and reconfigure itself to service requests
        exc = None
        for _i in range(5):
            time.sleep(2)
            try:
                self.clients[0].run_command(
                    ['curl', '--fail', self.acme_server])
                break
            except Exception as e:
                exc = e
        else:
            raise exc

    def test_centralize_acme_enable(self):
        """Test if ACME enable on replica if enabled on master"""
        status = check_acme_status(self.replicas[0], 'enabled')
        assert status == 'enabled'

    ###############
    # Certbot tests
    ###############

    @pytest.mark.skipif(skip_certbot_tests, reason='certbot not available')
    def test_certbot_register(self):
        certbot_register(self.clients[0], self.acme_server)

    @pytest.mark.skipif(skip_certbot_tests, reason='certbot not available')
    def test_certbot_certonly_standalone(self):
        certbot_standalone_cert(self.clients[0], self.acme_server)

    @pytest.mark.skipif(skip_certbot_tests, reason='certbot not available')
    def test_certbot_revoke(self):
        # Assume previous certonly operation succeeded.
        # Read certificate to learn serial number.
        cert_path = \
            f'/etc/letsencrypt/live/{self.clients[0].hostname}/cert.pem'
        data = self.clients[0].get_file_contents(cert_path)
        cert = x509.load_pem_x509_certificate(data, backend=default_backend())

        # revoke cert via ACME
        self.clients[0].run_command(
            [
                'certbot',
                '--server', self.acme_server,
                'revoke',
                '--cert-name', self.clients[0].hostname,
                '--delete-after-revoke',
            ],
        )

        # check cert is revoked (kinit already performed)
        result = self.master.run_command(
            ['ipa', 'cert-show', str(cert.serial_number), '--raw']
        )
        assert 'revocation_reason:' in result.stdout_text

    @pytest.mark.skipif(skip_certbot_tests, reason='certbot not available')
    def test_certbot_dns(self):
        # Assume previous revoke operation succeeded and cert was deleted.
        # We can now request a new certificate.

        # Get a cert from ACME service using dns-01 challenge and Certbot's
        # standalone HTTP server mode
        self.clients[0].run_command([
            'certbot',
            '--server', self.acme_server,
            'certonly',
            '--non-interactive',
            '--domain', self.clients[0].hostname,
            '--preferred-challenges', 'dns',
            '--manual',
            '--manual-public-ip-logging-ok',
            '--manual-auth-hook', CERTBOT_DNS_IPA_SCRIPT,
            '--manual-cleanup-hook', CERTBOT_DNS_IPA_SCRIPT,
        ])

    ##############
    # mod_md tests
    ##############

    @pytest.mark.skipif(skip_mod_md_tests, reason='mod_md not available')
    def test_mod_md(self):
        if get_selinux_status(self.clients[0]):
            # mod_md requires its own SELinux policy to grant perms to
            # maintaining ACME registration and cert state.
            raise pytest.skip("SELinux is enabled, this will fail")
        # write config
        self.clients[0].run_command(['mkdir', '-p', '/etc/httpd/conf.d'])
        self.clients[0].run_command(['mkdir', '-p', '/etc/httpd/md'])
        self.clients[0].put_file_contents(
            '/etc/httpd/conf.d/md.conf',
            '\n'.join([
                f'MDCertificateAuthority {self.acme_server}',
                'MDCertificateAgreement accepted',
                'MDStoreDir  /etc/httpd/md',
                f'MDomain {self.clients[0].hostname}',
                '<VirtualHost *:443>',
                f'    ServerName {self.clients[0].hostname}',
                '    SSLEngine on',
                '</VirtualHost>\n',
            ]),
        )

        # To check for successful cert issuance means knowing how mod_md
        # stores certificates, or looking for specific log messages.
        # If the thing we are inspecting changes, the test will break.
        # So I prefer a conservative sleep.
        #
        self.clients[0].run_command(['systemctl', 'restart', 'httpd'])
        time.sleep(15)

        # We expect mod_md has acquired the certificate by now.
        # Perform a graceful restart to begin using the cert.
        # (If mod_md ever learns to start using newly acquired
        # certificates /without/ the second restart, then both
        # of these sleeps can be replaced by "loop until good".)
        #
        self.clients[0].run_command(['systemctl', 'reload', 'httpd'])
        time.sleep(3)

        # HTTPS request from server to client (should succeed)
        self.master.run_command(
            ['curl', f'https://{self.clients[0].hostname}'])

        # clean-up
        self.clients[0].run_command(['rm', '-rf', '/etc/httpd/md'])
        self.clients[0].run_command(['rm', '-f', '/etc/httpd/conf.d/md.conf'])

    ######################
    # Disable ACME service
    ######################

    def test_disable_acme_service(self):
        """
        Disable ACME service again, and observe that it no longer services
        requests.

        """
        self.master.run_command(['ipa-acme-manage', 'disable'])

        # wait a short time for Dogtag ACME service to observe config
        # change and reconfigure itself to no longer service requests
        time.sleep(3)

        # should fail now
        self.clients[0].run_command(
            ['curl', '--fail', self.acme_server],
            ok_returncode=22,
        )

    def test_centralize_acme_disable(self):
        """Test if ACME disable on replica if disabled on master"""
        status = check_acme_status(self.replicas[0], 'disabled')
        assert status == 'disabled'

    @server_install_teardown
    def test_third_party_certs(self):
        """Require ipa-ca SAN on replacement web certificates"""

        self.master.run_command(['ipa-acme-manage', 'enable'])

        self.create_pkcs12('ca1/server')
        self.prepare_cacert('ca1')

        # Re-install the existing Apache certificate that has a SAN to
        # verify that it will be accepted.
        pin = self.master.get_file_contents(
            paths.HTTPD_PASSWD_FILE_FMT.format(host=self.master.hostname)
        )
        result = self.certinstall(
            certfile=paths.HTTPD_CERT_FILE,
            keyfile=paths.HTTPD_KEY_FILE,
            pin=pin
        )
        assert result.returncode == 0

        # Install using a 3rd party cert with a missing SAN for ipa-ca
        # which should be rejected.
        result = self.certinstall()
        assert result.returncode == 1

        self.master.run_command(['ipa-acme-manage', 'disable'])

        # Install using a 3rd party cert with a missing SAN for ipa-ca
        # which should be ok since ACME is disabled.
        result = self.certinstall()
        assert result.returncode == 0

        # Enable ACME which should fail since the Apache cert lacks the SAN
        result = self.master.run_command(['ipa-acme-manage', 'enable'],
                                         raiseonerr=False)
        assert result.returncode == 1
        assert "invalid 'certificate'" in result.stderr_text


class TestACMECALess(IntegrationTest):
    """Test to check the CA less replica setup"""
    num_replicas = 1
    num_clients = 0

    @pytest.fixture
    def test_setup_teardown(self):
        tasks.install_master(self.master, setup_dns=True)

        tasks.install_replica(self.master, self.replicas[0], setup_ca=False)

        yield

        tasks.uninstall_replica(self.master, self.replicas[0])
        tasks.uninstall_master(self.master)

    def test_caless_to_cafull_replica(self, test_setup_teardown):
        """Test ACME is enabled on CA-less replica when converted to CA-full

        Deployment where one server is deployed as CA-less, when converted
        to CA full, should have ACME enabled by default.

        related: https://pagure.io/freeipa/issue/8524
        """
        tasks.kinit_admin(self.master)
        # enable acme on master
        self.master.run_command(['ipa-acme-manage', 'enable'])

        # check status of acme server on master
        status = check_acme_status(self.master, 'enabled')
        assert status == 'enabled'

        tasks.kinit_admin(self.replicas[0])
        # check status of acme on replica, result: CA is not installed
        result = self.replicas[0].run_command(['ipa-acme-manage', 'status'],
                                              raiseonerr=False)
        assert result.returncode == 3

        # Install CA on replica
        tasks.install_ca(self.replicas[0])

        # check acme status, should be enabled now
        status = check_acme_status(self.replicas[0], 'enabled')
        assert status == 'enabled'

        # disable acme on replica
        self.replicas[0].run_command(['ipa-acme-manage', 'disable'])

        # check acme status on master, should be disabled
        status = check_acme_status(self.master, 'disabled')
        assert status == 'disabled'

    def test_enable_caless_to_cafull_replica(self, test_setup_teardown):
        """Test ACME with CA-less replica when converted to CA-full

        Deployment have one ca-less replica and ACME is not enabled.
        After converting ca-less replica to ca-full, ACME can be
        enabled or disabled.

        related: https://pagure.io/freeipa/issue/8524
        """
        tasks.kinit_admin(self.master)

        # check status of acme server on master
        status = check_acme_status(self.master, 'disabled')
        assert status == 'disabled'

        tasks.kinit_admin(self.replicas[0])
        # check status of acme on replica, result: CA is not installed
        result = self.replicas[0].run_command(['ipa-acme-manage', 'status'],
                                              raiseonerr=False)
        assert result.returncode == 3

        # Install CA on replica
        tasks.install_ca(self.replicas[0])

        # check acme status on replica, should not throw error
        status = check_acme_status(self.replicas[0], 'disabled')
        assert status == 'disabled'

        # enable acme on replica
        self.replicas[0].run_command(['ipa-acme-manage', 'enable'])

        # check acme status on master
        status = check_acme_status(self.master, 'enabled')
        assert status == 'enabled'

        # check acme status on replica
        status = check_acme_status(self.replicas[0], 'enabled')
        assert status == 'enabled'

        # disable acme on master
        self.master.run_command(['ipa-acme-manage', 'disable'])

        # check acme status on replica, should be disabled
        status = check_acme_status(self.replicas[0], 'disabled')
        assert status == 'disabled'


class TestACMEwithExternalCA(TestACME):
    """Test the FreeIPA ACME service with external CA"""

    num_replicas = 1
    num_clients = 1

    @classmethod
    def install(cls, mh):
        super(TestACMEwithExternalCA, cls).install(mh)

        # install master with external-ca
        result = install_server_external_ca_step1(cls.master)
        assert result.returncode == 0
        root_ca_fname, ipa_ca_fname = tasks.sign_ca_and_transport(
            cls.master, paths.ROOT_IPA_CSR, ROOT_CA, IPA_CA
        )

        install_server_external_ca_step2(
            cls.master, ipa_ca_fname, root_ca_fname
        )
        tasks.kinit_admin(cls.master)

        tasks.install_client(cls.master, cls.clients[0])
        tasks.install_replica(cls.master, cls.replicas[0])


class TestACMERenew(IntegrationTest):

    num_clients = 1

    @classmethod
    def install(cls, mh):

        # install packages before client install in case of IPA DNS problems
        cls.acme_server = prepare_acme_client(cls.master, cls.clients[0])

        tasks.install_master(cls.master, setup_dns=True)
        tasks.install_client(cls.master, cls.clients[0])

    @pytest.fixture
    def issue_and_expire_cert(self):
        """Fixture to expire cert by moving date past expiry of acme cert"""
        # enable the ACME service on master
        self.master.run_command(['ipa-acme-manage', 'enable'])

        # register the account with certbot
        certbot_register(self.clients[0], self.acme_server)

        # request a standalone acme cert
        certbot_standalone_cert(self.clients[0], self.acme_server)

        # move system date to expire acme cert
        for host in self.clients[0], self.master:
            tasks.kdestroy_all(host)
            tasks.move_date(host, 'stop', '+90days')

        tasks.get_kdcinfo(host)
        # Note raiseonerr=False:
        # the assert is located after kdcinfo retrieval.
        result = host.run_command(
            "KRB5_TRACE=/dev/stdout kinit admin",
            stdin_text='{0}\n{0}\n{0}\n'.format(
                self.clients[0].config.admin_password
            ),
            raiseonerr=False
        )
        # Retrieve kdc.$REALM after the password change, just in case SSSD
        # domain status flipped to online during the password change.
        tasks.get_kdcinfo(host)
        assert result.returncode == 0

        yield

        # move back date
        for host in self.clients[0], self.master:
            tasks.kdestroy_all(host)
            tasks.move_date(host, 'start', '-90days')
            tasks.kinit_admin(host)

    @pytest.mark.skipif(skip_certbot_tests, reason='certbot not available')
    def test_renew(self, issue_and_expire_cert):
        """Test if ACME renews the issued cert with cerbot

        This test is to check if ACME certificate renews upon
        reaching expiry

        related: https://pagure.io/freeipa/issue/4751
        """
        data = self.clients[0].get_file_contents(
            f'/etc/letsencrypt/live/{self.clients[0].hostname}/cert.pem'
        )
        cert = x509.load_pem_x509_certificate(data, backend=default_backend())
        initial_expiry = cert.not_valid_after

        self.clients[0].run_command(['certbot', 'renew'])

        data = self.clients[0].get_file_contents(
            f'/etc/letsencrypt/live/{self.clients[0].hostname}/cert.pem'
        )
        cert = x509.load_pem_x509_certificate(data, backend=default_backend())
        renewed_expiry = cert.not_valid_after

        assert initial_expiry != renewed_expiry



#
# Copyright (C) 2019  FreeIPA Contributors see COPYING for license
#

import os

import pytest

MOD_NAME = "test_module_{}"
FUNC_NAME = "test_func_{}"
MODS_NUM = 5


@pytest.fixture
def ipatestdir(testdir, monkeypatch):
    """
    Create MODS_NUM test modules within testdir/ipatests.
    Each module contains 1 test function.
    Patch PYTHONPATH with created package path to override the system's
    ipatests
    """
    ipatests_dir = testdir.mkpydir("ipatests")
    for i in range(MODS_NUM):
        ipatests_dir.join("{}.py".format(MOD_NAME.format(i))).write(
            "def {}(): pass".format(FUNC_NAME.format(i)))

    python_path = os.pathsep.join(
        filter(None, [str(testdir.tmpdir), os.environ.get("PYTHONPATH", "")]))
    monkeypatch.setenv("PYTHONPATH", python_path)

    def run_ipa_tests(*args):
        cmdargs = ["ipa-run-tests", "-v"] + list(args)
        return testdir.run(*cmdargs, timeout=60)

    testdir.run_ipa_tests = run_ipa_tests
    return testdir


def test_ipa_run_tests_basic(ipatestdir):
    """
    Run ipa-run-tests with default arguments
    """
    result = ipatestdir.run_ipa_tests()
    assert result.ret == 0
    result.assert_outcomes(passed=MODS_NUM)
    for mod_num in range(MODS_NUM):
        result.stdout.fnmatch_lines(["*{mod}.py::{func} PASSED*".format(
            mod=MOD_NAME.format(mod_num),
            func=FUNC_NAME.format(mod_num))])


def test_ipa_run_tests_glob1(ipatestdir):
    """
    Run ipa-run-tests using glob patterns to collect tests
    """
    result = ipatestdir.run_ipa_tests("{mod}".format(
        mod="test_modul[!E]?[0-5]*"))
    assert result.ret == 0
    result.assert_outcomes(passed=MODS_NUM)
    for mod_num in range(MODS_NUM):
        result.stdout.fnmatch_lines(["*{mod}.py::{func} PASSED*".format(
            mod=MOD_NAME.format(mod_num),
            func=FUNC_NAME.format(mod_num))])


def test_ipa_run_tests_glob2(ipatestdir):
    """
    Run ipa-run-tests using glob patterns to collect tests
    """
    result = ipatestdir.run_ipa_tests("{mod}".format(
        mod="test_module_{0,1}*"))
    assert result.ret == 0
    result.assert_outcomes(passed=2)
    for mod_num in range(2):
        result.stdout.fnmatch_lines(["*{mod}.py::{func} PASSED*".format(
            mod=MOD_NAME.format(mod_num),
            func=FUNC_NAME.format(mod_num))])


def test_ipa_run_tests_specific_nodeid(ipatestdir):
    """
    Run ipa-run-tests using nodeid to collect test
    """
    mod_num = 0
    result = ipatestdir.run_ipa_tests("{mod}.py::{func}".format(
        mod=MOD_NAME.format(mod_num),
        func=FUNC_NAME.format(mod_num)))
    assert result.ret == 0
    result.assert_outcomes(passed=1)
    result.stdout.fnmatch_lines(["*{mod}.py::{func} PASSED*".format(
        mod=MOD_NAME.format(mod_num),
        func=FUNC_NAME.format(mod_num))])


@pytest.mark.parametrize(
    "expr",
    [["-k", "not {func}".format(func=FUNC_NAME.format(0))],
     ["-k not {func}".format(func=FUNC_NAME.format(0))]])
def test_ipa_run_tests_expression(ipatestdir, expr):
    """
    Run ipa-run-tests using expression
    """
    result = ipatestdir.run_ipa_tests(*expr)
    assert result.ret == 0
    result.assert_outcomes(passed=4)
    for mod_num in range(1, MODS_NUM):
        result.stdout.fnmatch_lines(["*{mod}.py::{func} PASSED*".format(
            mod=MOD_NAME.format(mod_num),
            func=FUNC_NAME.format(mod_num))])


def test_ipa_run_tests_empty_expression(ipatestdir):
    """
    Run ipa-run-tests using an empty expression.
    Expected result: all tests should pass.
    """
    result = ipatestdir.run_ipa_tests('-k', '')
    assert result.ret == 0
    result.assert_outcomes(passed=5)
    for mod_num in range(0, MODS_NUM):
        result.stdout.fnmatch_lines(["*{mod}.py::{func} PASSED*".format(
            mod=MOD_NAME.format(mod_num),
            func=FUNC_NAME.format(mod_num))])


def test_ipa_run_tests_ignore_basic(ipatestdir):
    """
    Run ipa-run-tests ignoring one test module
    """
    result = ipatestdir.run_ipa_tests(
        "--ignore", "{mod}.py".format(mod=MOD_NAME.format(0)),
        "--ignore", "{mod}.py".format(mod=MOD_NAME.format(1)),
    )
    assert result.ret == 0
    result.assert_outcomes(passed=MODS_NUM - 2)
    for mod_num in range(2, MODS_NUM):
        result.stdout.fnmatch_lines(["*{mod}.py::{func} PASSED*".format(
            mod=MOD_NAME.format(mod_num),
            func=FUNC_NAME.format(mod_num))])


def test_ipa_run_tests_defaultargs(ipatestdir):
    """
    Checking the ipa-run-tests defaults:
    * cachedir
    * rootdir
    """
    mod_num = 0
    result = ipatestdir.run_ipa_tests("{mod}.py::{func}".format(
        mod=MOD_NAME.format(mod_num),
        func=FUNC_NAME.format(mod_num)))
    assert result.ret == 0
    result.assert_outcomes(passed=1)
    result.stdout.re_match_lines([
        "^cachedir: {cachedir}$".format(
            cachedir=os.path.join(os.getcwd(), ".pytest_cache")),
        "^rootdir: {rootdir}([,].*)?$".format(
            rootdir=os.path.join(str(ipatestdir.tmpdir), "ipatests"))
    ])


def test_ipa_run_tests_confcutdir(ipatestdir):
    """
    Checking the ipa-run-tests defaults:
    * confcutdir
    """
    mod_num = 0
    ipatestdir.makeconftest("import somenotexistedpackage")
    result = ipatestdir.run_ipa_tests("{mod}.py::{func}".format(
        mod=MOD_NAME.format(mod_num),
        func=FUNC_NAME.format(mod_num)))
    assert result.ret == 0
    result.assert_outcomes(passed=1)
    result.stdout.fnmatch_lines(["*{mod}.py::{func} PASSED*".format(
        mod=MOD_NAME.format(mod_num),
        func=FUNC_NAME.format(mod_num))])



# Copyright (C) 2015-2017  Custodia Project Contributors - see LICENSE file
from __future__ import absolute_import

import argparse
import os


class AbsFileType(argparse.FileType):
    """argparse file type with absolute path
    """
    def __call__(self, string):
        if string != '-':
            string = os.path.abspath(string)
        return super(AbsFileType, self).__call__(string)


class ConfigfileAction(argparse.Action):
    """Default action handler for configfile
    """
    default_path = '/etc/custodia/custodia.conf'
    default_instance = '/etc/custodia/{instance}.conf'

    def __call__(self, parser, namespace, values, option_string=None):
        if values is None:
            if namespace.instance is not None:
                values = self.default_instance.format(
                    instance=namespace.instance
                )
            else:
                values = self.default_path
            values = self.type(values)
        setattr(namespace, self.dest, values)


def instance_name(string):
    """Check for valid instance name
    """
    invalid = ':/@'
    if set(string).intersection(invalid):
        msg = 'Invalid instance name {}'.format(string)
        raise argparse.ArgumentTypeError(msg)
    return string


default_argparser = argparse.ArgumentParser(
    prog='custodia',
    description='Custodia server'
)
default_argparser.add_argument(
    '--debug',
    action='store_true',
    help='Debug mode'
)
default_argparser.add_argument(
    '--instance',
    type=instance_name,
    help='Instance name',
    default=None
)
default_argparser.add_argument(
    'configfile',
    nargs='?',
    action=ConfigfileAction,
    type=AbsFileType('r'),
    help=('Path to custodia server config (default: '
          '/etc/custodia/{instance}/custodia.conf)'),
)


def parse_args(args=None, argparser=None):
    if argparser is None:
        argparser = default_argparser

    # namespace with default values
    namespace = argparse.Namespace(
        debug=False,
        instance=None,
    )

    return argparser.parse_args(args, namespace)



#
# Copyright (C) 2019  FreeIPA Contributors see COPYING for license
#

"""Warns about xunit/unittest/nose tests.

FreeIPA is a rather old project and hereby includes all the most
famous in the past and present Python test idioms. Of course,
this is difficult to play around all of them. For now, the runner
of the IPA's test suite is Pytest.

Pytest supports xunit style setups, unittest, nose tests. But this
support is limited and may be dropped in the future releases.
Worst of all is that the mixing of various test frameworks results
in weird conflicts and of course, is not widely tested. In other
words, there is a big risk. To eliminate this risk and to not pin
Pytest to 3.x branch IPA's tests were refactored.

This plugin is intended to issue warnings on collecting tests,
which employ unittest/nose frameworks or xunit style.

To treat these warnings as errors it's enough to run Pytest with:

-W error:'xunit style is deprecated':pytest.PytestIPADeprecationWarning

"""
from unittest import TestCase

from ipatests.conftest import PytestIPADeprecationWarning

forbidden_module_scopes = [
    'setup_module',
    'setup_function',
    'teardown_module',
    'teardown_function',
]

forbidden_class_scopes = [
    'setup_class',
    'setup_method',
    'teardown_class',
    'teardown_method',
]


def pytest_collection_finish(session):
    for item in session.items:
        cls = getattr(item, 'cls', None)
        if cls is not None and issubclass(cls, TestCase):
            item.warn(PytestIPADeprecationWarning(
                "unittest is deprecated in favour of fixtures style"))
            continue

        def xunit_depr_warn(item, attr, names):
            for n in names:
                obj = getattr(item, attr, None)
                method = getattr(obj, n, None)
                fixtured = hasattr(method, '__pytest_wrapped__')
                if method is not None and not fixtured:
                    item.warn(PytestIPADeprecationWarning(
                        "xunit style is deprecated in favour of "
                        "fixtures style"))

        xunit_depr_warn(item, 'module', forbidden_module_scopes)
        xunit_depr_warn(item, 'cls', forbidden_class_scopes)



#
# Copyright (C) 2020  FreeIPA Contributors see COPYING for license
#
"""DNS forwarder and systemd-resolve1 helpers
"""
import ipaddress
import logging
import os
import socket

import dbus

from ipaplatform.paths import paths
from ipapython.dnsutil import get_ipa_resolver


logger = logging.getLogger(__name__)


_SYSTEMD_RESOLV_CONF = {
    "/run/systemd/resolve/stub-resolv.conf",
    "/run/systemd/resolve/resolv.conf",
    "/lib/systemd/resolv.conf",
    "/usr/lib/systemd/resolv.conf",
}

_DBUS_RESOLVE1_NAME = "org.freedesktop.resolve1"
_DBUS_RESOLVE1_PATH = "/org/freedesktop/resolve1"
_DBUS_RESOLVE1_MANAGER_IF = "org.freedesktop.resolve1.Manager"
_DBUS_PROPERTY_IF = "org.freedesktop.DBus.Properties"

# netlink interface index for resolve1 global settings and loopback
IFINDEX_GLOBAL = 0
IFINDEX_LOOPBACK = 1


def detect_resolve1_resolv_conf():
    """Detect if /etc/resolv.conf is managed by systemd-resolved

    See man(5) NetworkManager.conf
    """
    try:
        dest = os.readlink(paths.RESOLV_CONF)
    except OSError:
        # not a link
        return False
    # convert path relative to /etc/resolv.conf to abs path
    dest = os.path.normpath(
        os.path.join(os.path.dirname(paths.RESOLV_CONF), dest)
    )
    return dest in _SYSTEMD_RESOLV_CONF


def get_resolve1_nameservers(*, with_ifindex=False):
    """Get list of DNS nameservers from systemd-resolved

    :return: list of tuples (ifindex, ipaddress_obj)
    """
    bus = dbus.SystemBus()
    try:
        resolve1 = bus.get_object(_DBUS_RESOLVE1_NAME, _DBUS_RESOLVE1_PATH)
        prop_if = dbus.Interface(resolve1, _DBUS_PROPERTY_IF)
        dns_prop = prop_if.Get(_DBUS_RESOLVE1_MANAGER_IF, "DNSEx")
    finally:
        bus.close()

    results = []
    for ifindex, af, dns_arr, port, sniname in dns_prop:
        if port not in {0, 53} or sniname:
            # non-default port, non-standard port, or SNI name configuration
            # for DNS over TLS, e.g. 1.2.3.4:9953#example.com
            continue
        # convert packed format to IPAddress object (like inet_ntop)
        if af == socket.AF_INET:
            dnsip = ipaddress.IPv4Address(bytes(dns_arr))
        elif af == socket.AF_INET6:
            dnsip = ipaddress.IPv6Address(bytes(dns_arr))
        else:
            # neither IPv4 nor IPv6
            continue
        if with_ifindex:
            # netlink interface index, see socket.if_nameindex()
            ifindex = int(ifindex)
            results.append((ifindex, dnsip))
        else:
            results.append(dnsip)

    return results


def get_dnspython_nameservers(*, with_ifindex=False):
    """Get list of DNS nameservers from dnspython

    On Linux dnspython parses /etc/resolv.conf for us

    :return: list of tuples (ifindex, ipaddress_obj)
    """
    results = []
    for nameserver in get_ipa_resolver().nameservers:
        nameserver = ipaddress.ip_address(nameserver)
        if with_ifindex:
            results.append((IFINDEX_GLOBAL, nameserver))
        else:
            results.append(nameserver)
    return results


def get_nameservers():
    """Get list of unique, non-loopback DNS nameservers

    :return: list of strings
    """
    if detect_resolve1_resolv_conf():
        logger.debug(
            "systemd-resolved detected, fetching nameservers from D-Bus"
        )
        nameservers = get_resolve1_nameservers(with_ifindex=True)
    else:
        logger.debug(
            "systemd-resolved not detected, parsing %s", paths.RESOLV_CONF
        )
        nameservers = get_dnspython_nameservers(with_ifindex=True)

    logger.debug("Detected nameservers: %r", nameservers)

    result = []
    seen = set()
    for ifindex, ip in nameservers:
        # unique entries
        if ip in seen:
            continue
        seen.add(ip)
        # skip loopback
        if ifindex == IFINDEX_LOOPBACK or ip.is_loopback:
            continue
        result.append(str(ip))

    logger.debug("Use nameservers %r", result)

    return result


if __name__ == "__main__":
    from pprint import pprint

    print("systemd-resolved detected:", detect_resolve1_resolv_conf())
    print("Interfaces:", socket.if_nameindex())
    print("dnspython nameservers:")
    pprint(get_dnspython_nameservers(with_ifindex=True))
    print("resolve1 nameservers:")
    try:
        pprint(get_resolve1_nameservers(with_ifindex=True))
    except Exception as e:
        print(e)
    print("nameservers:", get_nameservers())



import pytest
import re
from contextlib import contextmanager

from ipatests.pytest_ipa.integration import tasks
from ipatests.test_integration.base import IntegrationTest


class TestResolverManager(IntegrationTest):
    topology = 'line'
    num_clients = 1
    invalid_resolver = '2.3.4.5'

    @classmethod
    def install(cls, mh):
        test_record = 'test1234'
        cls.client = cls.clients[0]
        cls.test_record = '{}.{}'.format(test_record, cls.master.domain.name)
        cls.test_record_address = '1.2.3.4'

        if cls.domain_level is not None:
            domain_level = cls.domain_level
        else:
            domain_level = cls.master.config.domain_level
        tasks.install_topo(cls.topology, cls.master, [], [], domain_level)
        tasks.kinit_admin(cls.master)
        cls.master.run_command([
            'ipa', 'dnsrecord-add', cls.master.domain.name,
            test_record,
            '--a-ip-address={}'.format(cls.test_record_address)])

    def is_resolver_operational(self):
        res = self.client.run_command(['dig', '+short', 'redhat.com'])
        return re.match(r'\d{1,3}.\d{1,3}.\d{1,3}.\d{1,3}',
                        res.stdout_text.strip())

    def is_ipa_dns_used(self):
        res = self.client.run_command(['dig', '+short', self.test_record])
        output = res.stdout_text.strip()
        error = res.stderr_text.strip()
        if output == self.test_record_address:
            return True
        if output == '' and error == '':
            return False
        raise Exception('Unexpected result of dig command')

    def check_ipa_name_resolution_fails(self):
        res = self.client.run_command(['dig', '+short', self.test_record],
                                      ok_returncode=9)
        assert 'connection timed out' in res.stdout_text

    @contextmanager
    def start_end_checks(self):
        assert not self.client.resolver.has_backups()
        assert self.is_resolver_operational()
        assert not self.is_ipa_dns_used()
        yield
        assert self.is_resolver_operational()
        assert not self.is_ipa_dns_used()
        assert not self.client.resolver.has_backups()

    def test_ipa_dns_not_used_by_default(self):
        assert self.is_resolver_operational()
        assert not self.is_ipa_dns_used()

    def test_changing_config_without_backup_not_allowed(self):
        with pytest.raises(Exception, match='without backup'):
            self.client.resolver.setup_resolver(self.master.ip)

    def test_change_resolver(self):
        with self.start_end_checks():
            self.client.resolver.backup()
            self.client.resolver.setup_resolver(self.master.ip)
            assert self.is_resolver_operational()
            assert self.is_ipa_dns_used()
            self.client.resolver.restore()

    def test_nested_change_resolver(self):
        with self.start_end_checks():
            self.client.resolver.backup()
            self.client.resolver.setup_resolver(self.master.ip)

            assert self.is_resolver_operational()
            assert self.is_ipa_dns_used()

            self.client.resolver.backup()
            self.client.resolver.setup_resolver(self.invalid_resolver)
            self.check_ipa_name_resolution_fails()
            self.client.resolver.restore()

            assert self.is_resolver_operational()
            assert self.is_ipa_dns_used()

            self.client.resolver.restore()

    def test_nested_change_resolver_with_context(self):
        with self.start_end_checks():
            self.client.resolver.backup()
            self.client.resolver.setup_resolver(self.master.ip)
            assert self.is_resolver_operational()
            assert self.is_ipa_dns_used()

            with self.client.resolver:
                self.client.resolver.setup_resolver(self.invalid_resolver)
                self.check_ipa_name_resolution_fails()

            self.client.resolver.restore()

    def test_repeated_changing_resolver(self):
        with self.start_end_checks():
            self.client.resolver.backup()
            self.client.resolver.setup_resolver(self.master.ip)
            assert self.is_resolver_operational()
            assert self.is_ipa_dns_used()

            self.client.resolver.setup_resolver(self.invalid_resolver)
            self.check_ipa_name_resolution_fails()

            self.client.resolver.setup_resolver(self.master.ip)
            assert self.is_resolver_operational()
            assert self.is_ipa_dns_used()

            self.client.resolver.restore()

    @pytest.mark.parametrize('reverse', [True, False])
    def test_multiple_resolvers(self, reverse):
        resolvers = [self.invalid_resolver, self.master.ip]
        if reverse:
            resolvers.reverse()

        with self.start_end_checks():
            self.client.resolver.backup()
            self.client.resolver.setup_resolver(resolvers)
            assert self.is_resolver_operational()
            assert self.is_ipa_dns_used()
            self.client.resolver.restore()

    @classmethod
    def uninstall(cls, mh):
        tasks.uninstall_master(cls.master)



#
# Copyright (C) 2019  FreeIPA Contributors see COPYING for license
#

# ipa-cert-fix performs the following steps:
#
# 1. Confirm running as root (AdminTool.validate_options does this)
#
# 2. Confirm that DS is up.
#
# 3. Determine which of following certs (if any) need renewing
#     - IPA RA
#     - Apache HTTPS
#     - 389 LDAPS
#     - Kerberos KDC (PKINIT)
#
# 4. Execute `pki-server cert-fix` with relevant options,
#    including `--extra-cert SERIAL` for each cert from #3.
#
# 5. Print details of renewed certificates.
#
# 6. Install renewed certs from #3 in relevant places
#
# 7. ipactl restart

from __future__ import print_function, absolute_import

import base64
from cryptography import x509 as crypto_x509
from cryptography.hazmat.backends import default_backend
import datetime
from enum import Enum
import logging
import shutil

from ipalib import api
from ipalib import x509
from ipalib.facts import is_ipa_configured
from ipalib.install import certmonger
from ipaplatform.paths import paths
from ipapython.admintool import AdminTool
from ipapython.certdb import NSSDatabase, EMPTY_TRUST_FLAGS
from ipapython.dn import DN
from ipapython.ipaldap import realm_to_serverid
from ipaserver.install import ca, cainstance, dsinstance
from ipaserver.install.certs import is_ipa_issued_cert
from ipapython import directivesetter
from ipapython import ipautil

msg = """
                          WARNING

ipa-cert-fix is intended for recovery when expired certificates
prevent the normal operation of IPA.  It should ONLY be used
in such scenarios, and backup of the system, especially certificates
and keys, is STRONGLY RECOMMENDED.

"""

renewal_note = """
Note: Monitor the certmonger-initiated renewal of
certificates after ipa-cert-fix and wait for its completion before
any other administrative task.
"""

RENEWED_CERT_PATH_TEMPLATE = "/etc/pki/pki-tomcat/certs/{}-renewed.crt"

logger = logging.getLogger(__name__)


cert_nicknames = {
    'sslserver': 'Server-Cert cert-pki-ca',
    'subsystem': 'subsystemCert cert-pki-ca',
    'ca_ocsp_signing': 'ocspSigningCert cert-pki-ca',
    'ca_audit_signing': 'auditSigningCert cert-pki-ca',
    'kra_transport': 'transportCert cert-pki-kra',
    'kra_storage': 'storageCert cert-pki-kra',
    'kra_audit_signing': 'auditSigningCert cert-pki-kra',
}

class IPACertType(Enum):
    IPARA = "IPA RA"
    HTTPS = "Apache HTTPS"
    LDAPS = "LDAP"
    KDC = "KDC"


class IPACertFix(AdminTool):
    command_name = "ipa-cert-fix"
    usage = "%prog"
    description = "Renew expired certificates."

    def validate_options(self):
        super(IPACertFix, self).validate_options(needs_root=True)

    def run(self):
        if not is_ipa_configured():
            print("IPA is not configured.")
            return 2

        if not cainstance.is_ca_installed_locally():
            print("CA is not installed on this server.")
            return 1

        try:
            ipautil.run(['pki-server', 'cert-fix', '--help'], raiseonerr=True)
        except ipautil.CalledProcessError:
            print(
                "The 'pki-server cert-fix' command is not available; "
                "cannot proceed."
            )
            return 1

        api.bootstrap(in_server=True, confdir=paths.ETC_IPA)
        api.finalize()

        if not dsinstance.is_ds_running(realm_to_serverid(api.env.realm)):
            print(
                "The LDAP server is not running; cannot proceed."
            )
            return 1

        api.Backend.ldap2.connect()  # ensure DS is up

        subject_base = dsinstance.DsInstance().find_subject_base()
        if not subject_base:
            raise RuntimeError("Cannot determine certificate subject base.")

        ca_subject_dn = ca.lookup_ca_subject(api, subject_base)

        now = datetime.datetime.now() + datetime.timedelta(weeks=2)
        certs, extra_certs, non_renewed = expired_certs(now)

        if not certs and not extra_certs:
            print("Nothing to do.")
            return 0

        print(msg)

        print_intentions(certs, extra_certs, non_renewed)

        response = ipautil.user_input('Enter "yes" to proceed')
        if response.lower() != 'yes':
            print("Not proceeding.")
            return 0
        print("Proceeding.")

        try:
            fix_certreq_directives(certs)
            run_cert_fix(certs, extra_certs)
        except ipautil.CalledProcessError:
            if any(
                x[0] is IPACertType.LDAPS
                for x in extra_certs + non_renewed
            ):
                # The DS cert was expired.  This will cause 'pki-server
                # cert-fix' to fail at the final restart, and return nonzero.
                # So this exception *might* be OK to ignore.
                #
                # If 'pki-server cert-fix' has written new certificates
                # corresponding to all the extra_certs, then ignore the
                # CalledProcessError and proceed to installing the IPA-specific
                # certs.  Otherwise re-raise.
                if check_renewed_ipa_certs(extra_certs):
                    pass
                else:
                    raise
            else:
                raise  # otherwise re-raise

        replicate_dogtag_certs(subject_base, ca_subject_dn, certs)
        install_ipa_certs(subject_base, ca_subject_dn, extra_certs)

        if any(x[0] != 'sslserver' for x in certs) \
                or any(x[0] is IPACertType.IPARA for x in extra_certs):
            # we renewed a "shared" certificate, therefore we must
            # become the renewal master
            print("Becoming renewal master.")
            cainstance.CAInstance().set_renewal_master()

        print("Restarting IPA")
        ipautil.run(['ipactl', 'restart'], raiseonerr=True)

        print(renewal_note)
        return 0


def expired_certs(now):
    expired_ipa, non_renew_ipa = expired_ipa_certs(now)
    return expired_dogtag_certs(now), expired_ipa, non_renew_ipa


def expired_dogtag_certs(now):
    """
    Determine which Dogtag certs are expired, or close to expiry.

    Return a list of (cert_id, cert) pairs.

    """
    certs = []
    db = NSSDatabase(nssdir=paths.PKI_TOMCAT_ALIAS_DIR)

    for certid, nickname in cert_nicknames.items():
        try:
            cert = db.get_cert(nickname)
        except RuntimeError:
            pass  # unfortunately certdb doesn't give us a better exception
        else:
            if cert.not_valid_after <= now:
                certs.append((certid, cert))

    return certs


def expired_ipa_certs(now):
    """
    Determine which IPA certs are expired, or close to expiry.

    Return a list of (IPACertType, cert) pairs.

    """
    certs = []
    non_renewed = []

    # IPA RA
    cert = x509.load_certificate_from_file(paths.RA_AGENT_PEM)
    if cert.not_valid_after <= now:
        certs.append((IPACertType.IPARA, cert))

    # Apache HTTPD
    cert = x509.load_certificate_from_file(paths.HTTPD_CERT_FILE)
    if cert.not_valid_after <= now:
        if not is_ipa_issued_cert(api, cert):
            non_renewed.append((IPACertType.HTTPS, cert))
        else:
            certs.append((IPACertType.HTTPS, cert))

    # LDAPS
    serverid = realm_to_serverid(api.env.realm)
    ds = dsinstance.DsInstance(realm_name=api.env.realm)
    ds_dbdir = dsinstance.config_dirname(serverid)
    ds_nickname = ds.get_server_cert_nickname(serverid)
    db = NSSDatabase(nssdir=ds_dbdir)
    cert = db.get_cert(ds_nickname)
    if cert.not_valid_after <= now:
        if not is_ipa_issued_cert(api, cert):
            non_renewed.append((IPACertType.LDAPS, cert))
        else:
            certs.append((IPACertType.LDAPS, cert))

    # KDC
    cert = x509.load_certificate_from_file(paths.KDC_CERT)
    if cert.not_valid_after <= now:
        if not is_ipa_issued_cert(api, cert):
            non_renewed.append((IPACertType.HTTPS, cert))
        else:
            certs.append((IPACertType.KDC, cert))

    return certs, non_renewed


def print_intentions(dogtag_certs, ipa_certs, non_renewed):
    print("The following certificates will be renewed:")
    print()

    for certid, cert in dogtag_certs:
        print_cert_info("Dogtag", certid, cert)

    for certtype, cert in ipa_certs:
        print_cert_info("IPA", certtype.value, cert)

    if non_renewed:
        print(
            "The following certificates will NOT be renewed because "
            "they were not issued by the IPA CA:"
        )
        print()

        for certtype, cert in non_renewed:
            print_cert_info("IPA", certtype.value, cert)


def print_cert_info(context, desc, cert):
    print("{} {} certificate:".format(context, desc))
    print("  Subject: {}".format(DN(cert.subject)))
    print("  Serial:  {}".format(cert.serial_number))
    print("  Expires: {}".format(cert.not_valid_after))
    print()


def get_csr_from_certmonger(nickname):
    """
    Get the csr for the provided nickname by asking certmonger.

    Returns the csr in ASCII format without the header/footer in a single line
    or None if not found.
    """
    criteria = {
        'cert-database': paths.PKI_TOMCAT_ALIAS_DIR,
        'cert-nickname': nickname,
    }

    id = certmonger.get_request_id(criteria)
    if id:
        csr = certmonger.get_request_value(id, "csr")
        if csr:
            try:
                # Make sure the value can be parsed as valid CSR
                csr_obj = crypto_x509.load_pem_x509_csr(
                    csr.encode('ascii'), default_backend())
                val = base64.b64encode(csr_obj.public_bytes(x509.Encoding.DER))
                return val.decode('ascii')
            except Exception as e:
                # Fallthrough and return None
                logger.debug("Unable to get CSR from certmonger: %s", e)
    return None


def fix_certreq_directives(certs):
    """
    For all the certs to be fixed, ensure that the corresponding CSR is found
    in PKI config file, or try to get the CSR from certmonger.
    """
    directives = {
        'auditSigningCert cert-pki-ca': ('ca.audit_signing.certreq',
                                         paths.CA_CS_CFG_PATH),
        'ocspSigningCert cert-pki-ca': ('ca.ocsp_signing.certreq',
                                        paths.CA_CS_CFG_PATH),
        'subsystemCert cert-pki-ca': ('ca.subsystem.certreq',
                                      paths.CA_CS_CFG_PATH),
        'Server-Cert cert-pki-ca': ('ca.sslserver.certreq',
                                    paths.CA_CS_CFG_PATH),
        'auditSigningCert cert-pki-kra': ('kra.audit_signing.certreq',
                                          paths.KRA_CS_CFG_PATH),
        'storageCert cert-pki-kra': ('kra.storage.certreq',
                                     paths.KRA_CS_CFG_PATH),
        'transportCert cert-pki-kra': ('kra.transport.certreq',
                                       paths.KRA_CS_CFG_PATH),
    }

    # pki-server cert-fix needs to find the CSR in the subsystem config file
    # otherwise it will fail
    # For each cert to be fixed, check that the CSR is present or
    # get it from certmonger
    for (certid, _cert) in certs:
        # Check if the directive is set in the config file
        nickname = cert_nicknames[certid]
        (directive, cfg_path) = directives[nickname]
        if directivesetter.get_directive(cfg_path, directive, '=') is None:
            # The CSR is missing, try to get it from certmonger
            csr = get_csr_from_certmonger(nickname)
            if csr:
                # Update the directive
                directivesetter.set_directive(cfg_path, directive, csr,
                                              quotes=False, separator='=')


def run_cert_fix(certs, extra_certs):
    ldapi_path = (
        paths.SLAPD_INSTANCE_SOCKET_TEMPLATE
        % '-'.join(api.env.realm.split('.'))
    )
    cmd = [
        'pki-server',
        'cert-fix',
        '--ldapi-socket', ldapi_path,
        '--agent-uid', 'ipara',
    ]
    for certid, _cert in certs:
        cmd.extend(['--cert', certid])
    for _certtype, cert in extra_certs:
        cmd.extend(['--extra-cert', str(cert.serial_number)])
    ipautil.run(cmd, raiseonerr=True)


def replicate_dogtag_certs(subject_base, ca_subject_dn, certs):
    for certid, _oldcert in certs:
        cert_path = "/etc/pki/pki-tomcat/certs/{}.crt".format(certid)
        cert = x509.load_certificate_from_file(cert_path)
        print_cert_info("Renewed Dogtag", certid, cert)
        replicate_cert(subject_base, ca_subject_dn, cert)


def check_renewed_ipa_certs(certs):
    """
    Check whether all expected IPA-specific certs (extra_certs) were renewed
    successfully.

    For now this subroutine just checks that the files that we expect
    ``pki-server cert-fix`` to have written do exist and contain an X.509
    certificate.

    Return ``True`` if everything seems to be as expected, otherwise ``False``.

    """
    for _certtype, oldcert in certs:
        cert_path = RENEWED_CERT_PATH_TEMPLATE.format(oldcert.serial_number)
        try:
            x509.load_certificate_from_file(cert_path)
        except (IOError, ValueError):
            return False

    return True


def install_ipa_certs(subject_base, ca_subject_dn, certs):
    """Print details and install renewed IPA certificates."""
    for certtype, oldcert in certs:
        cert_path = RENEWED_CERT_PATH_TEMPLATE.format(oldcert.serial_number)
        cert = x509.load_certificate_from_file(cert_path)
        print_cert_info("Renewed IPA", certtype.value, cert)

        if certtype is IPACertType.IPARA:
            shutil.copyfile(cert_path, paths.RA_AGENT_PEM)
            cainstance.update_people_entry(cert)
            replicate_cert(subject_base, ca_subject_dn, cert)
        elif certtype is IPACertType.HTTPS:
            shutil.copyfile(cert_path, paths.HTTPD_CERT_FILE)
        elif certtype is IPACertType.LDAPS:
            serverid = realm_to_serverid(api.env.realm)
            ds = dsinstance.DsInstance(realm_name=api.env.realm)
            ds_dbdir = dsinstance.config_dirname(serverid)
            db = NSSDatabase(nssdir=ds_dbdir)
            ds_nickname = ds.get_server_cert_nickname(serverid)
            db.delete_cert(ds_nickname)
            db.import_pem_cert(ds_nickname, EMPTY_TRUST_FLAGS, cert_path)
        elif certtype is IPACertType.KDC:
            shutil.copyfile(cert_path, paths.KDC_CERT)


def replicate_cert(subject_base, ca_subject_dn, cert):
    nickname = cainstance.get_ca_renewal_nickname(
        subject_base, ca_subject_dn, DN(cert.subject))
    if nickname:
        cainstance.update_ca_renewal_entry(api.Backend.ldap2, nickname, cert)



#
# Copyright (C) 2020  FreeIPA Contributors see COPYING for license
#

"""
Facts about the installation
"""

import logging
import os
from . import sysrestore
from ipaplatform.paths import paths

logger = logging.getLogger(__name__)

# Used to determine install status
IPA_MODULES = [
    'httpd', 'kadmin', 'dirsrv', 'pki-tomcatd', 'install', 'krb5kdc', 'named']


def is_ipa_configured():
    """
    Use the state to determine if IPA has been configured.
    """
    sstore = sysrestore.StateFile(paths.SYSRESTORE)
    if sstore.has_state('installation'):
        return sstore.get_state('installation', 'complete')

    # Fall back to older method in case this is an existing installation

    installed = False

    fstore = sysrestore.FileStore(paths.SYSRESTORE)

    for module in IPA_MODULES:
        if sstore.has_state(module):
            logger.debug('%s is configured', module)
            installed = True
        else:
            logger.debug('%s is not configured', module)

    if fstore.has_files():
        logger.debug('filestore has files')
        installed = True
    else:
        logger.debug('filestore is tracking no files')

    return installed


def is_ipa_client_configured(on_master=False):
    """
    Consider IPA client not installed if nothing is backed up
    and default.conf file does not exist. If on_master is set to True,
    the existence of default.conf file is not taken into consideration,
    since it has been already created by ipa-server-install.
    """
    fstore = sysrestore.FileStore(paths.IPA_CLIENT_SYSRESTORE)
    statestore = sysrestore.StateFile(paths.IPA_CLIENT_SYSRESTORE)

    installed = statestore.get_state('installation', 'complete')
    if installed is not None:
        return installed

    # Fall back to the old detection

    installed = (
        fstore.has_files() or (
            not on_master and os.path.exists(paths.IPA_DEFAULT_CONF)
        )
    )

    return installed



#
# Copyright (C) 2020 FreeIPA Contributors, see COPYING for license
#

"""
This module contains default SUSE OS family-specific implementations of
system tasks.
"""

import logging

from ipaplatform.paths import paths
from ipaplatform.redhat.tasks import RedHatTaskNamespace

logger = logging.getLogger(__name__)


class SuseTaskNamespace(RedHatTaskNamespace):
    def restore_context(self, filepath, force=False):
        pass  # FIXME: Implement after libexec move

    def check_selinux_status(self, restorecon=paths.RESTORECON):
        pass  # FIXME: Implement after libexec move

    def set_nisdomain(self, nisdomain):
        nis_variable = "NETCONFIG_NIS_STATIC_DOMAIN"
        try:
            with open(paths.SYSCONF_NETWORK, "r") as f:
                content = [
                    line
                    for line in f
                    if not line.strip().upper().startswith(nis_variable)
                ]
        except IOError:
            content = []

        content.append("{}={}\n".format(nis_variable, nisdomain))

        with open(paths.SYSCONF_NETWORK, "w") as f:
            f.writelines(content)

    def set_selinux_booleans(self, required_settings, backup_func=None):
        return False  # FIXME: Implement after libexec move


tasks = SuseTaskNamespace()



#
# Copyright (C) 2015  FreeIPA Contributors see COPYING for license
#

"""
Module provides tests which testing ability of various subsystems to be
installed.
"""

from __future__ import absolute_import

import os
import re
import textwrap
import time
from datetime import datetime, timedelta

import pytest
from cryptography.hazmat.primitives import hashes
from cryptography import x509 as crypto_x509

from ipalib import x509
from ipalib.constants import DOMAIN_LEVEL_0, KRA_TRACKING_REQS
from ipalib.constants import IPA_CA_RECORD
from ipalib.constants import ALLOWED_NETBIOS_CHARS
from ipalib.sysrestore import SYSRESTORE_STATEFILE, SYSRESTORE_INDEXFILE
from ipapython.dn import DN
from ipaplatform.constants import constants
from ipaplatform.osinfo import osinfo
from ipaplatform.paths import paths
from ipaplatform.tasks import tasks as platformtasks
from ipapython import ipautil
from ipatests.pytest_ipa.integration import tasks
from ipatests.pytest_ipa.integration.env_config import get_global_config
from ipatests.test_integration.base import IntegrationTest
from ipatests.test_integration.test_caless import CALessBase, ipa_certs_cleanup
from ipaplatform import services


config = get_global_config()


def create_broken_resolv_conf(master):
    # Force a broken resolv.conf to simulate a bad response to
    # reverse zone lookups
    master.resolver.backup()
    master.resolver.setup_resolver('127.0.0.2')


def server_install_setup(func):
    def wrapped(*args):
        master = args[0].master
        create_broken_resolv_conf(master)
        try:
            func(*args)
        finally:
            tasks.uninstall_master(master, clean=False)
            ipa_certs_cleanup(master)
    return wrapped


@pytest.fixture
def server_cleanup(request):
    """
    Fixture to uninstall ipa server before and after the test
    """
    host = request.cls.master
    tasks.uninstall_master(host)
    yield
    tasks.uninstall_master(host)


def create_netbios_name(host):
    """
    Create a NetBIOS name based on the provided host
    """
    netbios = ''.join(
        c for c in host.domain.name.split('.')[0].upper() \
        if c in ALLOWED_NETBIOS_CHARS
    )[:15]
    return netbios


class InstallTestBase1(IntegrationTest):

    num_replicas = 3
    topology = 'star'
    master_with_dns = False

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=cls.master_with_dns)

    def test_replica0_ca_less_install(self):
        tasks.install_replica(
            self.master, self.replicas[0], setup_ca=False,
            nameservers='master' if self.master_with_dns else None)

    def test_replica0_ipa_ca_install(self):
        tasks.install_ca(self.replicas[0])

    def test_replica0_ipa_kra_install(self):
        tasks.install_kra(self.replicas[0], first_instance=True)

    def test_replica0_ipa_dns_install(self):
        tasks.install_dns(self.replicas[0])

    def test_replica1_with_ca_install(self):
        tasks.install_replica(
            self.master, self.replicas[1], setup_ca=True,
            nameservers='master' if self.master_with_dns else None)

    def test_replica1_ipa_kra_install(self):
        tasks.install_kra(self.replicas[1])

    def test_replica1_ipa_dns_install(self):
        tasks.install_dns(self.replicas[1])

    def test_replica2_with_ca_kra_install(self):
        tasks.install_replica(
            self.master, self.replicas[2], setup_ca=True, setup_kra=True,
            nameservers='master' if self.master_with_dns else None)

    def test_replica2_ipa_dns_install(self):
        tasks.install_dns(self.replicas[2])


class InstallTestBase2(IntegrationTest):

    num_replicas = 3
    topology = 'star'
    master_with_dns = False

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=cls.master_with_dns)

    def test_replica1_with_ca_dns_install(self):
        tasks.install_replica(
            self.master, self.replicas[1], setup_ca=True, setup_dns=True,
            nameservers='master' if self.master_with_dns else None)

    def test_replica1_ipa_kra_install(self):
        tasks.install_kra(self.replicas[1])

    def test_replica2_with_dns_install(self):
        tasks.install_replica(
            self.master, self.replicas[2], setup_ca=False, setup_dns=True,
            nameservers='master' if self.master_with_dns else None)

    def test_replica2_ipa_ca_install(self):
        tasks.install_ca(self.replicas[2])

    def test_replica2_ipa_kra_install(self):
        tasks.install_kra(self.replicas[2])


class ADTrustInstallTestBase(IntegrationTest):
    """
    Base test for builtin AD trust installation im combination with other
    components
    """
    num_replicas = 2
    topology = 'star'
    master_with_dns = False

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=cls.master_with_dns)

    def install_replica(self, replica, **kwargs):
        tasks.install_replica(self.master, replica, setup_adtrust=True,
                              **kwargs)

    def test_replica0_only_adtrust(self):
        self.install_replica(
            self.replicas[0], setup_ca=False,
            nameservers='master' if self.master_with_dns else None)

    def test_replica1_all_components_adtrust(self):
        self.install_replica(
            self.replicas[1], setup_ca=True,
            nameservers='master' if self.master_with_dns else None)


##
# Master X Replicas installation tests
##

class TestInstallWithCA1(InstallTestBase1):
    master_with_dns = False

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=cls.master_with_dns)

    @pytest.mark.skipif(config.domain_level == DOMAIN_LEVEL_0,
                        reason='does not work on DOMAIN_LEVEL_0 by design')
    def test_replica1_ipa_kra_install(self):
        super(TestInstallWithCA1, self).test_replica1_ipa_kra_install()

    @pytest.mark.skipif(config.domain_level == DOMAIN_LEVEL_0,
                        reason='does not work on DOMAIN_LEVEL_0 by design')
    def test_replica2_with_ca_kra_install(self):
        super(TestInstallWithCA1, self).test_replica2_with_ca_kra_install()

    @pytest.mark.skipif(config.domain_level == DOMAIN_LEVEL_0,
                        reason='does not work on DOMAIN_LEVEL_0 by design')
    def test_replica2_ipa_dns_install(self):
        super(TestInstallWithCA1, self).test_replica2_ipa_dns_install()

    def test_install_with_bad_ldap_conf(self):
        """
        Test a client install with a non standard ldap.config
        https://pagure.io/freeipa/issue/7418
        """
        ldap_conf = paths.OPENLDAP_LDAP_CONF
        base_dn = self.master.domain.basedn
        client = self.replicas[0]
        tasks.uninstall_replica(self.master, client)
        expected_msg1 = "contains deprecated and unsupported " \
                        "entries: HOST, PORT"
        file_backup = client.get_file_contents(ldap_conf, encoding='utf-8')
        constants = "URI ldaps://{}\nBASE {}\nHOST {}\nPORT 636".format(
            self.master.hostname, base_dn,
            self.master.hostname)
        modifications = "{}\n{}".format(file_backup, constants)
        client.put_file_contents(paths.OPENLDAP_LDAP_CONF, modifications)
        result = client.run_command(['ipa-client-install', '-U',
                                     '--domain', client.domain.name,
                                     '--realm', client.domain.realm,
                                     '-p', client.config.admin_name,
                                     '-w', client.config.admin_password,
                                     '--server', self.master.hostname],
                                    raiseonerr=False)
        assert expected_msg1 in result.stderr_text
        client.put_file_contents(ldap_conf, file_backup)


class TestInstallWithCA2(InstallTestBase2):
    master_with_dns = False

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=cls.master_with_dns)

    @pytest.mark.skipif(config.domain_level == DOMAIN_LEVEL_0,
                        reason='does not work on DOMAIN_LEVEL_0 by design')
    def test_replica1_ipa_kra_install(self):
        super(TestInstallWithCA2, self).test_replica1_ipa_kra_install()

    @pytest.mark.skipif(config.domain_level == DOMAIN_LEVEL_0,
                        reason='does not work on DOMAIN_LEVEL_0 by design')
    def test_replica2_ipa_kra_install(self):
        super(TestInstallWithCA2, self).test_replica2_ipa_kra_install()


class TestInstallCA(IntegrationTest):
    """
    Tests for CA installation on a replica
    """

    num_replicas = 2

    @classmethod
    def install(cls, mh):
        cls.master.put_file_contents(
            os.path.join(paths.IPA_CCACHES, 'foo'),
            'somerandomstring'
        )
        cls.master.run_command(
            ['mkdir', os.path.join(paths.IPA_CCACHES, 'bar')]
        )
        tasks.install_master(cls.master, setup_dns=False)

    def test_ccaches_cleanup(self):
        """
        The IPA ccaches directory is cleaned up on install. Verify
        that the file we created is now gone.
        """
        assert os.path.exists(os.path.join(paths.IPA_CCACHES, 'foo')) is False
        assert os.path.exists(os.path.join(paths.IPA_CCACHES, 'bar')) is False

    def test_replica_ca_install_with_no_host_dns(self):
        """
        Test for ipa-ca-install --no-host-dns on a replica
        """

        tasks.install_replica(self.master, self.replicas[0], setup_ca=False)
        tasks.install_ca(self.replicas[0], extra_args=["--no-host-dns"])

    def test_replica_ca_install_with_skip_schema_check(self):
        """
        Test for ipa-ca-install --skip-schema-check on a replica
        """

        tasks.install_replica(self.master, self.replicas[1], setup_ca=False)
        tasks.install_ca(self.replicas[1], extra_args=["--skip-schema-check"])

    def test_certmonger_reads_token_HSM(self):
        """Test if certmonger reads the token in HSM

        This is to ensure added HSM support for FreeIPA. This test adds
        certificate with sofhsm token and checks if certmonger is tracking
        it.

        related : https://pagure.io/certmonger/issue/125
        """
        test_service = 'test/%s' % self.master.hostname
        pkcs_passwd = 'Secret123'
        pin = '123456'
        noisefile = '/tmp/noisefile'
        self.master.put_file_contents(noisefile, os.urandom(64))

        tasks.kinit_admin(self.master)
        tasks.install_dns(self.master)
        self.master.run_command(['ipa', 'service-add', test_service])

        # create a csr
        cmd_args = ['certutil', '-d', paths.NSS_DB_DIR, '-R', '-a',
                    '-o', '/root/ipa.csr',
                    '-s', "CN=%s" % self.master.hostname,
                    '-z', noisefile]
        self.master.run_command(cmd_args)

        # request certificate
        cmd_args = ['ipa', 'cert-request', '--principal', test_service,
                    '--certificate-out', '/root/test.pem', '/root/ipa.csr']
        self.master.run_command(cmd_args)

        # adding trust flag
        cmd_args = ['certutil', '-A', '-d', paths.NSS_DB_DIR, '-n',
                    'test', '-a', '-i', '/root/test.pem', '-t', 'u,u,u']
        self.master.run_command(cmd_args)

        # export pkcs12 file
        cmd_args = ['pk12util', '-o', '/root/test.p12',
                    '-d', paths.NSS_DB_DIR, '-n', 'test', '-W', pkcs_passwd]
        self.master.run_command(cmd_args)

        # add softhsm lib
        cmd_args = ['modutil', '-dbdir', paths.NSS_DB_DIR, '-add',
                    'softhsm', '-libfile', '/usr/lib64/softhsm/libsofthsm.so']
        self.master.run_command(cmd_args, stdin_text="\n\n")

        # create a token
        cmd_args = ['softhsm2-util', '--init-token', '--label', 'test',
                    '--pin', pin, '--so-pin', pin, '--free']
        self.master.run_command(cmd_args)

        self.master.run_command(['softhsm2-util', '--show-slots'])

        cmd_args = ['certutil', '-F', '-d', paths.NSS_DB_DIR, '-n', 'test']
        self.master.run_command(cmd_args)

        cmd_args = ['pk12util', '-i', '/root/test.p12',
                    '-d', paths.NSS_DB_DIR, '-h', 'test',
                    '-W', pkcs_passwd, '-K', pin]
        self.master.run_command(cmd_args)

        cmd_args = ['certutil', '-A', '-d', paths.NSS_DB_DIR, '-n', 'IPA CA',
                    '-t', 'CT,,', '-a', '-i', paths.IPA_CA_CRT]
        self.master.run_command(cmd_args)

        # validate the certificate
        self.master.put_file_contents('/root/pinfile', pin)
        cmd_args = ['certutil', '-V', '-u', 'V', '-e', '-d', paths.NSS_DB_DIR,
                    '-h', 'test', '-n', 'test:test', '-f', '/root/pinfile']
        result = self.master.run_command(cmd_args)
        assert 'certificate is valid' in result.stdout_text

        # add certificate tracking to certmonger
        cmd_args = ['ipa-getcert', 'start-tracking', '-d', paths.NSS_DB_DIR,
                    '-n', 'test', '-t', 'test', '-P', pin,
                    '-K', test_service]
        result = self.master.run_command(cmd_args)
        request_id = re.findall(r'\d+', result.stdout_text)

        # check if certificate is tracked by certmonger
        status = tasks.wait_for_request(self.master, request_id[0], 300)
        assert status == "MONITORING"

        # ensure if key and token are re-usable
        cmd_args = ['getcert', 'resubmit', '-i', request_id[0]]
        self.master.run_command(cmd_args)

        status = tasks.wait_for_request(self.master, request_id[0], 300)
        assert status == "MONITORING"

    def test_ipa_ca_crt_permissions(self):
        """Verify that /etc/ipa/ca.cert is mode 0644 root:root"""
        result = self.master.run_command(
            ["/usr/bin/stat", "-c", "%U:%G:%a", paths.IPA_CA_CRT]
        )
        out = str(result.stdout_text.strip())
        (owner, group, mode) = out.split(':')
        assert mode == "644"
        assert owner == "root"
        assert group == "root"

    def test_cert_install_with_IPA_issued_cert(self):
        """
        Test replacing an IPA-issued server cert

        ipa-server-certinstall can replace the web and LDAP certs.
        A slightly different code path is taken when the replacement
        certs are issued by IPA. Exercise that path by replacing the
        web cert with itself.
        """
        self.master.run_command(['cp', '-p', paths.HTTPD_CERT_FILE, '/tmp'])
        self.master.run_command(['cp', '-p', paths.HTTPD_KEY_FILE, '/tmp'])

        passwd = self.master.get_file_contents(
            paths.HTTPD_PASSWD_FILE_FMT.format(host=self.master.hostname)
        )
        self.master.run_command([
            'ipa-server-certinstall',
            '-p', self.master.config.dirman_password,
            '-w',
            '--pin', passwd,
            '/tmp/httpd.crt',
            '/tmp/httpd.key',
        ])

    def test_is_ipa_configured(self):
        """Verify that the old and new methods of is_ipa_installed works

           If there is an installation section then it is the status.

           If not then it will fall back to looking for configured
           services and files and use that for determination.
        """
        def set_installation_state(host, state):
            """
            Update the complete value in the installation section
            """
            host.run_command(
                ['python3', '-c',
                 'from ipalib.install import sysrestore; '
                 'from ipaplatform.paths import paths;'
                 'sstore = sysrestore.StateFile(paths.SYSRESTORE); '
                 'sstore.backup_state("installation", "complete", '
                 '{state})'.format(state=state)])

        def get_installation_state(host):
            """
            Retrieve the installation state from new install method
            """
            result = host.run_command(
                ['python3', '-c',
                 'from ipalib.install import sysrestore; '
                 'from ipaplatform.paths import paths;'
                 'sstore = sysrestore.StateFile(paths.SYSRESTORE); '
                 'print(sstore.get_state("installation", "complete"))'])
            return result.stdout_text.strip()  # a string

        # This comes from freeipa.spec and is used to determine whether
        # an upgrade is required.
        cmd = ['python3', '-c',
               'import sys; from ipalib import facts; sys.exit(0 '
               'if facts.is_ipa_configured() else 1);']

        # This will use the new method since this is a fresh install,
        # verify that it is true.
        self.master.run_command(cmd)
        assert get_installation_state(self.master) == 'True'

        # Set complete to False which should cause the command to fail
        # This tests the state of a failed or in-process installation.
        set_installation_state(self.master, False)
        result = self.master.run_command(cmd, raiseonerr=False)
        assert result.returncode == 1
        set_installation_state(self.master, True)

        # Tweak sysrestore.state to drop installation section
        self.master.run_command(
            ['sed','-i', r's/\[installation\]/\[badinstallation\]/',
             os.path.join(paths.SYSRESTORE, SYSRESTORE_STATEFILE)])

        # Re-run installation check and it should fall back to old method
        # and be successful.
        self.master.run_command(cmd)
        assert get_installation_state(self.master) == 'None'

        # Restore installation section.
        self.master.run_command(
            ['sed','-i', r's/\[badinstallation\]/\[installation\]/',
             os.path.join(paths.SYSRESTORE, SYSRESTORE_STATEFILE)])

        # Uninstall and confirm that the old method reports correctly
        # on uninstalled servers. It will exercise the old method since
        # there is no state.
        tasks.uninstall_master(self.master)

        # ensure there is no stale state
        result = self.master.run_command(r'test -f {}'.format(
            os.path.join(paths.SYSRESTORE, SYSRESTORE_STATEFILE)),
            raiseonerr=False
        )
        assert result.returncode == 1
        result = self.master.run_command(r'test -f {}'.format(
            os.path.join(paths.SYSRESTORE, SYSRESTORE_INDEXFILE)),
            raiseonerr=False
        )
        assert result.returncode == 1

        # Now run is_ipa_configured() and it should be False
        result = self.master.run_command(cmd, raiseonerr=False)
        assert result.returncode == 1


class TestInstallWithCA_KRA1(InstallTestBase1):
    master_with_dns = False

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=cls.master_with_dns,
                             setup_kra=True)

    def test_replica0_ipa_kra_install(self):
        tasks.install_kra(self.replicas[0], first_instance=False)


class TestInstallWithCA_KRA2(InstallTestBase2):
    master_with_dns = False

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=cls.master_with_dns,
                             setup_kra=True)


class TestInstallWithCA_DNS1(InstallTestBase1):
    master_with_dns = True

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=cls.master_with_dns)

    @pytest.mark.skipif(config.domain_level == DOMAIN_LEVEL_0,
                        reason='does not work on DOMAIN_LEVEL_0 by design')
    def test_replica1_ipa_kra_install(self):
        super(TestInstallWithCA_DNS1, self).test_replica1_ipa_kra_install()

    @pytest.mark.skipif(config.domain_level == DOMAIN_LEVEL_0,
                        reason='does not work on DOMAIN_LEVEL_0 by design')
    def test_replica2_with_ca_kra_install(self):
        super(TestInstallWithCA_DNS1, self).test_replica2_with_ca_kra_install()

    @pytest.mark.skipif(config.domain_level == DOMAIN_LEVEL_0,
                        reason='does not work on DOMAIN_LEVEL_0 by design')
    def test_replica2_ipa_dns_install(self):
        super(TestInstallWithCA_DNS1, self).test_replica2_ipa_dns_install()


class TestInstallWithCA_DNS2(InstallTestBase2):
    master_with_dns = True

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=cls.master_with_dns)

    @pytest.mark.skipif(config.domain_level == DOMAIN_LEVEL_0,
                        reason='does not work on DOMAIN_LEVEL_0 by design')
    def test_replica1_ipa_kra_install(self):
        super(TestInstallWithCA_DNS2, self).test_replica1_ipa_kra_install()

    @pytest.mark.skipif(config.domain_level == DOMAIN_LEVEL_0,
                        reason='does not work on DOMAIN_LEVEL_0 by design')
    def test_replica2_ipa_kra_install(self):
        super(TestInstallWithCA_DNS2, self).test_replica2_ipa_kra_install()


class TestInstallWithCA_DNS3(CALessBase):
    """
    Test an install with a bad DNS resolver configured to force a
    timeout trying to verify the existing zones. In the case of a reverse
    zone it is skipped unless --allow-zone-overlap is set regardless of
    the value of --auto-reverse. Confirm that --allow-zone-overlap
    lets the reverse zone be created.

    ticket 7239
    """

    @pytest.mark.xfail(
        osinfo.id == 'fedora' and osinfo.version_number >= (33,)
        and osinfo.version_number < (35,),
        reason='freeipa ticket 8700', strict=True)
    @server_install_setup
    def test_number_of_zones(self):
        """There should be two zones: one forward, one reverse"""

        self.create_pkcs12('ca1/server')
        self.prepare_cacert('ca1')

        self.install_server(extra_args=['--allow-zone-overlap'])

        result = self.master.run_command([
            'ipa', 'dnszone-find'])

        assert "in-addr.arpa." in result.stdout_text

        assert "returned 2" in result.stdout_text


class TestInstallWithCA_DNS4(CALessBase):
    """
    Test an install with a bad DNS resolver configured to force a
    timeout trying to verify the existing zones. In the case of a reverse
    zone it is skipped unless --allow-zone-overlap is set regardless of
    the value of --auto-reverse. Confirm that without --allow-reverse-zone
    only the forward zone is created.

    ticket 7239
    """

    @server_install_setup
    def test_number_of_zones(self):
        """There should be one zone, a forward because rev timed-out"""

        self.create_pkcs12('ca1/server')
        self.prepare_cacert('ca1')

        # no zone overlap by default
        self.install_server()

        result = self.master.run_command([
            'ipa', 'dnszone-find'])

        assert "in-addr.arpa." not in result.stdout_text

        assert "returned 1" in result.stdout_text


@pytest.mark.cs_acceptance
class TestInstallWithCA_KRA_DNS1(InstallTestBase1):
    master_with_dns = True

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=cls.master_with_dns,
                             setup_kra=True)

    def test_replica0_ipa_kra_install(self):
        tasks.install_kra(self.replicas[0], first_instance=False)


class TestInstallWithCA_KRA_DNS2(InstallTestBase2):
    master_with_dns = True

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=cls.master_with_dns,
                             setup_kra=True)


class TestADTrustInstall(ADTrustInstallTestBase):
    """
    Tests built-in AD trust installation in various combinations (see the base
    class for more details) against plain IPA master (no DNS, no KRA, no AD
    trust)
    """


class TestADTrustInstallWithDNS_KRA_ADTrust(ADTrustInstallTestBase):
    """
    Tests built-in AD trust installation in various combinations (see the base
    class for more details) against fully equipped (DNS, CA, KRA, ADtrust)
    master. Additional two test cases were added to test interplay including
    KRA installer
    """
    master_with_dns = True

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=cls.master_with_dns,
                             setup_kra=True, setup_adtrust=True)

    def test_replica1_all_components_adtrust(self):
        self.install_replica(self.replicas[1], setup_ca=True, setup_kra=True)


def get_pki_tomcatd_pid(host):
    pid = ''
    cmd = host.run_command(['systemctl', 'status', 'pki-tomcatd@pki-tomcat'])
    for line in cmd.stdout_text.split('\n'):
        if "Main PID" in line:
            pid = line.split()[2]
            break
    return(pid)

def get_ipa_services_pids(host):
    ipa_services_name = [
        "krb5kdc", "kadmin", "named", "httpd", "ipa-custodia",
        "pki_tomcatd", "ipa-dnskeysyncd"
    ]
    pids_of_ipa_services = {}
    for name in ipa_services_name:
        service_name = services.knownservices[name].systemd_name
        result = host.run_command(
            ["systemctl", "-p", "MainPID", "--value", "show", service_name]
        )
        pids_of_ipa_services[service_name] = int(result.stdout_text.strip())
    return pids_of_ipa_services


##
# Rest of master installation tests
##

class TestInstallMaster(IntegrationTest):

    num_replicas = 0

    @classmethod
    def install(cls, mh):
        pass

    def test_install_master(self):
        tasks.install_master(self.master, setup_dns=False)

    @pytest.mark.skip_if_platform(
        "debian", reason="This test hardcodes the httpd service name"
    )
    def test_smoke_test_for_debug_mode(self):
        """Test if an IPA server works in debug mode.
        Related: https://pagure.io/freeipa/issue/8891

        Note: this test hardcodes the "httpd" service name.
        """

        target_fname = paths.IPA_SERVER_CONF
        assert not self.master.transport.file_exists(target_fname)

        # set the IPA server in debug mode
        server_conf = "[global]\ndebug=True"
        self.master.put_file_contents(target_fname, server_conf)
        self.master.run_command(["systemctl", "restart", "httpd"])

        # smoke test in debug mode
        tasks.kdestroy_all(self.master)
        tasks.kinit_admin(self.master)
        self.master.run_command(["ipa", "user-show", "admin"])

        # rollback
        self.master.run_command(["rm", target_fname])
        self.master.run_command(["systemctl", "restart", "httpd"])

    def test_schema_compat_attribute_and_tree_disable(self):
        """Test if schema-compat-entry-attribute is set

        This is to ensure if said entry is set after installation.
        It also checks if compat tree is disable.

        related: https://pagure.io/freeipa/issue/8193
        """
        conn = self.master.ldap_connect()
        entry = conn.get_entry(DN(             # pylint: disable=no-member
            "cn=groups,cn=Schema Compatibility,cn=plugins,cn=config"))

        entry_list = list(entry['schema-compat-entry-attribute'])
        value = (r'ipaexternalmember=%deref_r('
                 '"member","ipaexternalmember")')
        assert value in entry_list
        assert 'schema-compat-lookup-nsswitch' not in entry_list

    def test_install_kra(self):
        tasks.install_kra(self.master, first_instance=True)

    def test_install_dns(self):
        tasks.install_dns(
            self.master,
            extra_args=['--dnssec-master', '--no-dnssec-validation']
        )

    def test_ipactl_restart_pki_tomcat(self):
        """ Test if ipactl restart restarts the pki-tomcatd

        Wrong logic was triggering the start instead of restart
        for pki-tomcatd. This test validates that restart
        called on pki-tomcat properly.

        related ticket : https://pagure.io/freeipa/issue/7927
        """
        # get process id of pki-tomcatd
        pki_pid = get_pki_tomcatd_pid(self.master)

        # check if pki-tomcad restarted
        cmd = self.master.run_command(['ipactl', 'restart'])
        assert "Restarting pki-tomcatd Service" in cmd.stdout_text

        # check if pid for pki-tomcad changed
        pki_pid_after_restart = get_pki_tomcatd_pid(self.master)
        assert pki_pid != pki_pid_after_restart

        # check if pki-tomcad restarted
        cmd = self.master.run_command(['ipactl', 'restart'])
        assert "Restarting pki-tomcatd Service" in cmd.stdout_text

        # check if pid for pki-tomcad changed
        pki_pid_after_restart_2 = get_pki_tomcatd_pid(self.master)
        assert pki_pid_after_restart != pki_pid_after_restart_2

    def test_ipactl_scenario_check(self):
        """ Test if ipactl starts services stopped by systemctl
        This will first check if all services are running then it will stop
        few service. After that it will restart all services and then check
        the status and pid of services.It will also compare pid after ipactl
        start and restart in case of start it will remain unchanged on the
        other hand in case of restart it will change.
        """
        # listing all services
        ipa_services_name = [
            "Directory", "krb5kdc", "kadmin", "named", "httpd", "ipa-custodia",
            "pki-tomcatd", "ipa-otpd", "ipa-dnskeysyncd"
        ]

        # checking the service status
        cmd = self.master.run_command(['ipactl', 'status'])
        for service in ipa_services_name:
            assert f"{service} Service: RUNNING" in cmd.stdout_text

        # stopping few services
        service_stop = ["krb5kdc", "kadmin", "httpd"]
        for service in service_stop:
            service_name = services.knownservices[service].systemd_name
            self.master.run_command(['systemctl', 'stop', service_name])

        # checking service status
        service_start = [
            svcs for svcs in ipa_services_name if svcs not in service_stop
        ]
        cmd = self.master.run_command(['ipactl', 'status'], raiseonerr=False)
        assert cmd.returncode == 3
        for service in service_start:
            assert f"{service} Service: RUNNING" in cmd.stdout_text
        for service in service_stop:
            assert f'{service} Service: STOPPED' in cmd.stdout_text

        # starting all services again
        self.master.run_command(['ipactl', 'start'])

        # checking service status
        cmd = self.master.run_command(['ipactl', 'status'])
        for service in ipa_services_name:
            assert f"{service} Service: RUNNING" in cmd.stdout_text

        # get process id of services
        ipa_services_pids = get_ipa_services_pids(self.master)

        # restarting all services again
        self.master.run_command(['ipactl', 'restart'])

        # checking service status
        cmd = self.master.run_command(['ipactl', 'status'])
        for service in ipa_services_name:
            assert f"{service} Service: RUNNING" in cmd.stdout_text

        # check if pid for services are different
        svcs_pids_after_restart = get_ipa_services_pids(self.master)
        assert ipa_services_pids != svcs_pids_after_restart

        # starting all services again
        self.master.run_command(['ipactl', 'start'])

        # checking service status
        cmd = self.master.run_command(['ipactl', 'status'])
        for service in ipa_services_name:
            assert f"{service} Service: RUNNING" in cmd.stdout_text

        # check if pid for services are same
        svcs_pids_after_start = get_ipa_services_pids(self.master)
        assert svcs_pids_after_restart == svcs_pids_after_start

    def test_WSGI_worker_process(self):
        """ Test if WSGI worker process count is set to 4

        related ticket : https://pagure.io/freeipa/issue/7587
        """
        # check process count in httpd conf file i.e expected string
        exp = b'WSGIDaemonProcess ipa processes=%d' % constants.WSGI_PROCESSES
        httpd_conf = self.master.get_file_contents(paths.HTTPD_IPA_CONF)
        assert exp in httpd_conf

        # check the process count
        cmd = self.master.run_command('ps -eF')
        wsgi_count = cmd.stdout_text.count('wsgi:ipa')
        assert constants.WSGI_PROCESSES == wsgi_count

    def test_error_for_yubikey(self):
        """ Test error when yubikey hardware not present

        In order to work with IPA and Yubikey, libyubikey is required.
        Before the fix, if yubikey added without having packages, it used to
        result in traceback. Now it the exception is handeled properly.
        It needs Yubikey hardware to make command successfull. This test
        just check of proper error thrown when hardware is not attached.

        related ticket : https://pagure.io/freeipa/issue/6979
        """
        # try to add yubikey to the user
        args = ['ipa', 'otptoken-add-yubikey', '--owner=admin']
        cmd = self.master.run_command(args, raiseonerr=False)
        assert cmd.returncode != 0
        exp_str = ("ipa: ERROR: No YubiKey found")
        assert exp_str in cmd.stderr_text

    def test_pki_certs(self):
        certs, keys = tasks.certutil_certs_keys(
            self.master,
            paths.PKI_TOMCAT_ALIAS_DIR,
            paths.PKI_TOMCAT_ALIAS_PWDFILE_TXT
        )

        expected_certs = {
            # CA
            'caSigningCert cert-pki-ca': 'CTu,Cu,Cu',
            'ocspSigningCert cert-pki-ca': 'u,u,u',
            'subsystemCert cert-pki-ca': 'u,u,u',
            'auditSigningCert cert-pki-ca': 'u,u,Pu',  # why P?
            # KRA
            'transportCert cert-pki-kra': 'u,u,u',
            'storageCert cert-pki-kra': 'u,u,u',
            'auditSigningCert cert-pki-kra': 'u,u,Pu',
            # server
            'Server-Cert cert-pki-ca': 'u,u,u',
        }
        assert certs == expected_certs
        assert len(certs) == len(keys)

        for nickname in sorted(certs):
            cert = tasks.certutil_fetch_cert(
                self.master,
                paths.PKI_TOMCAT_ALIAS_DIR,
                paths.PKI_TOMCAT_ALIAS_PWDFILE_TXT,
                nickname
            )
            key_size = cert.public_key().key_size
            if nickname == 'caSigningCert cert-pki-ca':
                assert key_size == 3072
            else:
                assert key_size == 2048
            assert cert.signature_hash_algorithm.name == hashes.SHA256.name

    def test_http_cert(self):
        """
        Test that HTTP certificate contains ipa-ca.$DOMAIN
        DNS name.

        """
        data = self.master.get_file_contents(paths.HTTPD_CERT_FILE)
        cert = x509.load_pem_x509_certificate(data)
        name = f'ipa-ca.{self.master.domain.name}'
        assert crypto_x509.DNSName(name) in cert.san_general_names

    def test_ipa_cert_in_store(self):
        """
        Test that IPA cert has been added to trust store.
        """

        assert "IPA CA" in self.master.run_command(
            ['trust', 'list'],
            raiseonerr=False).stdout_text

    def test_p11_kit_softhsm2(self):
        # check that p11-kit-proxy does not inject SoftHSM2
        result = self.master.run_command([
            "modutil", "-dbdir", paths.PKI_TOMCAT_ALIAS_DIR, "-list"
        ])
        assert "softhsm" not in result.stdout_text.lower()
        assert "opendnssec" not in result.stdout_text.lower()

    @pytest.mark.skipif(
        not platformtasks.is_selinux_enabled(),
        reason="Test needs SELinux enabled")
    def test_selinux_avcs(self):
        # Use journalctl instead of ausearch. The ausearch command is not
        # installed by default and journalctl gives us all AVCs.
        result = self.master.run_command([
            "journalctl", "--full", "--grep=AVC", "--since=yesterday"
        ], raiseonerr=False)
        avcs = list(
            line.strip() for line in result.stdout_text.split('\n')
            if "AVC avc:" in line
        )
        if avcs:
            print('\n'.join(avcs))
            # Use expected failure until all SELinux violations are fixed
            pytest.xfail("{} AVCs found".format(len(avcs)))

    def test_file_permissions(self):
        args = [
            "rpm", "-V",
            "python3-ipaclient",
            "python3-ipalib",
            "python3-ipaserver"
        ]

        if osinfo.id == 'fedora':
            args.extend([
                "freeipa-client",
                "freeipa-client-common",
                "freeipa-common",
                "freeipa-server",
                "freeipa-server-common",
                "freeipa-server-dns",
                "freeipa-server-trust-ad"
            ])
        else:
            args.extend([
                "ipa-client",
                "ipa-client-common",
                "ipa-common",
                "ipa-server",
                "ipa-server-common",
                "ipa-server-dns"
            ])

        result = self.master.run_command(args, raiseonerr=False)
        if result.returncode != 0:
            # Check the mode errors
            mode_warnings = re.findall(
                r"^.M.......  [cdglr ]+ (?P<filename>.*)$",
                result.stdout_text, re.MULTILINE)
            msg = "rpm -V found mode issues for the following files: {}"
            assert mode_warnings == [], msg.format(mode_warnings)
            # Check the owner errors
            user_warnings = re.findall(
                r"^.....U...  [cdglr ]+ (?P<filename>.*)$",
                result.stdout_text, re.MULTILINE)
            msg = "rpm -V found ownership issues for the following files: {}"
            assert user_warnings == [], msg.format(user_warnings)
            # Check the group errors
            group_warnings = re.findall(
                r"^......G..  [cdglr ]+ (?P<filename>.*)$",
                result.stdout_text, re.MULTILINE)
            msg = "rpm -V found group issues for the following files: {}"
            assert group_warnings == [], msg.format(group_warnings)

    def test_ds_disable_upgrade_hash(self):
        # Test case for https://pagure.io/freeipa/issue/8315
        # Disable password schema migration on LDAP bind
        result = tasks.ldapsearch_dm(
            self.master,
            "cn=config",
            ldap_args=["nsslapd-enable-upgrade-hash"],
            scope="base"
        )
        assert "nsslapd-enable-upgrade-hash: off" in result.stdout_text

    def test_ldbm_tuning(self):
        # check db-locks in new cn=bdb subentry (1.4.3+)
        result = tasks.ldapsearch_dm(
            self.master,
            "cn=bdb,cn=config,cn=ldbm database,cn=plugins,cn=config",
            ["nsslapd-db-locks"],
            scope="base"
        )
        assert "nsslapd-db-locks: 50000" in result.stdout_text

        # no db-locks configuration in old global entry
        result = tasks.ldapsearch_dm(
            self.master,
            "cn=config,cn=ldbm database,cn=plugins,cn=config",
            ["nsslapd-db-locks"],
            scope="base"
        )
        assert "nsslapd-db-locks" not in result.stdout_text

    def test_admin_root_alias_CVE_2020_10747(self):
        # Test for CVE-2020-10747 fix
        # https://bugzilla.redhat.com/show_bug.cgi?id=1810160
        rootprinc = "root@{}".format(self.master.domain.realm)
        result = self.master.run_command(["ipa", "user-show", "admin"])
        assert rootprinc in result.stdout_text

        result = self.master.run_command(
            ["ipa", "user-add", "root", "--first", "root", "--last", "root"],
            raiseonerr=False
        )
        assert result.returncode != 0
        assert 'user with name "root" already exists' in result.stderr_text

    def test_dirsrv_no_ssca(self):
        # verify that lib389 installer no longer creates self-signed CA
        result = self.master.run_command(
            ["stat", "/etc/dirsrv/ssca"],
            raiseonerr=False
        )
        assert result.returncode != 0

    def test_ipa_custodia_check(self):
        # check local key retrieval
        self.master.run_command(
            [paths.IPA_CUSTODIA_CHECK, self.master.hostname]
        )

    @pytest.mark.skipif(
        paths.SEMODULE is None, reason="test requires semodule command"
    )
    def test_ipa_selinux_policy(self):
        # check that freeipa-selinux's policy module is loaded and
        # not disabled
        result = self.master.run_command(
            [paths.SEMODULE, "-lfull"]
        )
        # prio module pp [disabled]
        # 100: default priority
        # 200: decentralized SELinux policy priority
        entries = {
            tuple(line.split())
            for line in result.stdout_text.split('\n')
            if line.strip()
        }
        assert ('200', 'ipa', 'pp') in entries

    def test_ipaca_no_redirect(self):
        """Test that ipa-ca.$DOMAIN does not redirect

           ipa-ca is a valid name for an IPA server. It should not
           require a redirect.

           CRL generation does not need to be enabled for this test.
           We aren't exactly testing that a CRL can be retrieved, just
           that the redirect doesn't happen.
        """

        def run_request(url, expected_stdout=None, expected_stderr=None):
            result = self.master.run_command(['curl', '-s', '-v', url])
            if expected_stdout:
                assert expected_stdout in result.stdout_text
            if expected_stderr:
                assert expected_stderr in result.stderr_text

        # CRL publishing on start-up is disabled so drop a file there
        crlfile = os.path.join(paths.PKI_CA_PUBLISH_DIR, 'MasterCRL.bin')
        self.master.put_file_contents(crlfile, 'secret')

        hosts = (
            f'{IPA_CA_RECORD}.{self.master.domain.name}',
            self.master.hostname,
        )

        # Positive tests. Both hosts can serve these.
        urls = (
            'http://{host}/ipa/crl/MasterCRL.bin',
            'http://{host}/ca/ocsp',
            'https://{host}/ca/admin/ca/getCertChain',
            'https://{host}/acme/',
        )
        for url in urls:
            for host in hosts:
                run_request(
                    url.format(host=host),
                    expected_stderr='HTTP/1.1 200'
                )

        # Negative tests. ipa-ca cannot serve these and will redirect and
        # test that existing redirect for unencrypted still works
        urls = (
            'http://{host}/',
            'http://{host}/ipa/json',
            'http://{carecord}.{domain}/ipa/json',
            'https://{carecord}.{domain}/ipa/json',
            'http://{carecord}.{domain}/ipa/config/ca.crt',
        )
        for url in urls:
            run_request(
                url.format(host=self.master.hostname,
                           domain=self.master.domain.name,
                           carecord=IPA_CA_RECORD),
                expected_stdout=f'href="https://{self.master.hostname}/'
            )

    def test_hostname_parameter(self, server_cleanup):
        """
        Test that --hostname parameter is respected in interactive mode.

        https://pagure.io/freeipa/issue/2692
        """
        original_hostname = self.master.hostname
        new_hostname = 'new.' + original_hostname
        # New hostname is added into /etc/hosts as the installer
        # is looking for it for resolution. Without it, an installation
        # fails with `Unable to resolve host name, check /etc/hosts or DNS
        # name resolution`.
        hosts = self.master.get_file_contents(paths.HOSTS, encoding='utf-8')
        new_hosts = hosts.replace(original_hostname, new_hostname)
        self.master.put_file_contents(paths.HOSTS, new_hosts)
        netbios = create_netbios_name(self.master)

        try:
            cmd = ['ipa-server-install', '--hostname', new_hostname]
            with self.master.spawn_expect(cmd) as e:
                e.expect_exact('Do you want to configure integrated '
                               'DNS (BIND)? [no]: ')
                e.sendline('no')
                e.expect_exact('Please confirm the domain name [{}]: '.format(
                    original_hostname  # DN is computed from new hostname
                ))
                e.sendline(self.master.domain.name)
                e.expect_exact('Please provide a realm name [{}]: '.format(
                    self.master.domain.realm
                ))
                e.sendline(self.master.domain.realm)
                e.expect_exact('Directory Manager password: ')
                e.sendline(self.master.config.dirman_password)
                e.expect_exact('Password (confirm): ')
                e.sendline(self.master.config.dirman_password)
                e.expect_exact('IPA admin password: ')
                e.sendline(self.master.config.admin_password)
                e.expect_exact('Password (confirm): ')
                e.sendline(self.master.config.admin_password)
                e.expect_exact('NetBIOS domain name [{}]: '.format(netbios))
                e.sendline(netbios)
                e.expect_exact('Do you want to configure chrony with '
                               'NTP server or pool address? [no]: ')
                e.sendline('no')
                e.expect_exact('Continue to configure the system '
                               'with these values? [no]: ')
                e.sendline('yes')
                e.expect_exit(ignore_remaining_output=True, timeout=720)
            hostname = self.master.run_command([
                'hostname', '-f']).stdout_text.strip()
            assert hostname == new_hostname
        finally:
            # no need to restore the hostname as the installer
            # does it during uninstallation
            self.master.put_file_contents(paths.HOSTS, hosts)

    def test_ad_subpackage_dependency(self, server_cleanup):
        """
        Test if the installer is not dependant on trust-ad package and
        succeeds even when AD subpackage is not installed

        https://pagure.io/freeipa/issue/4011
        """
        if osinfo.id == 'fedora':
            package_name = 'freeipa-server-trust-ad'
        else:
            package_name = 'ipa-server-trust-ad'
        reinstall = False
        if tasks.is_package_installed(self.master, package_name):
            tasks.uninstall_packages(self.master, [package_name])
            reinstall = True
        try:
            tasks.install_master(self.master)
        finally:
            if reinstall:
                tasks.install_packages(self.master, [package_name])

    def test_backup_of_cs_cfg_is_created(self, server_cleanup):
        """
        Test that the installer backs up CS.cfg configuration before it's
        being modified.

        https://pagure.io/freeipa/issue/4166
        """
        bcp_location = paths.CA_CS_CFG_PATH + '.ipabkp'
        original_cfg_content = None
        if self.master.transport.file_exists(paths.CA_CS_CFG_PATH):
            original_cfg_content = self.master.get_file_contents(
                paths.CA_CS_CFG_PATH, encoding='utf-8')

        time_before_install = int(self.master.run_command(
            ['date', '+%s']).stdout_text.strip())
        tasks.install_master(self.master)
        ipaserver_install_log = self.master.get_file_contents(
            paths.IPASERVER_INSTALL_LOG, encoding='utf-8')
        assert 'backing up CS.cfg' in ipaserver_install_log
        assert self.master.transport.file_exists(bcp_location)

        cfg_mod_time = int(self.master.run_command(
            ['stat', '-c', '%Y', paths.CA_CS_CFG_PATH]).stdout_text.strip())
        bcp_mod_time = int(self.master.run_command(
            ['stat', '-c', '%Y', bcp_location]).stdout_text.strip())

        # check if the backup file was modified/created during the installation
        # and before the original file was modified
        assert time_before_install <= bcp_mod_time <= cfg_mod_time
        bcp_cfg_content = self.master.get_file_contents(
            paths.CA_CS_CFG_PATH + '.ipabkp', encoding='utf-8')
        if original_cfg_content is not None:
            assert original_cfg_content == bcp_cfg_content


class TestInstallMasterKRA(IntegrationTest):

    num_replicas = 0

    @classmethod
    def install(cls, mh):
        pass

    def test_install_master(self):
        tasks.install_master(self.master, setup_dns=False, setup_kra=True)

    def test_install_dns(self):
        tasks.install_dns(self.master)

    def test_kra_certs_renewal(self):
        """
        Test that the KRA subsystem certificates renew properly
        """
        for nickname in KRA_TRACKING_REQS:
            cert = tasks.certutil_fetch_cert(
                self.master,
                paths.PKI_TOMCAT_ALIAS_DIR,
                paths.PKI_TOMCAT_ALIAS_PWDFILE_TXT,
                nickname
            )
            starting_serial = int(cert.serial_number)
            cmd_arg = [
                'ipa-getcert', 'resubmit', '-v', '-w',
                '-d', paths.PKI_TOMCAT_ALIAS_DIR,
                '-n', nickname,
            ]
            result = self.master.run_command(cmd_arg)
            request_id = re.findall(r'\d+', result.stdout_text)

            status = tasks.wait_for_request(self.master, request_id[0], 120)
            assert status == "MONITORING"

            cert = tasks.certutil_fetch_cert(
                self.master,
                paths.PKI_TOMCAT_ALIAS_DIR,
                paths.PKI_TOMCAT_ALIAS_PWDFILE_TXT,
                nickname
            )
            assert starting_serial != int(cert.serial_number)


class TestInstallMasterDNS(IntegrationTest):

    num_replicas = 0

    @classmethod
    def install(cls, mh):
        pass

    def test_install_master(self):
        tasks.install_master(
            self.master,
            setup_dns=True,
            extra_args=['--zonemgr', 'me@example.org'],
        )

        tasks.kinit_admin(self.master)
        result = self.master.run_command(
            ['ipa', 'dnszone-show', self.master.domain.name]
        ).stdout_text

        assert "Administrator e-mail address: me.example.org" in result

    def test_server_install_lock_bind_recursion(self):
        """Test if server installer lock Bind9 recursion

        This test is to check if recursion can be configured.
        It checks if newly added file /etc/named/ipa-ext.conf
        exists and /etc/named.conf should not have
        'allow-recursion { any; };'. It also checks if ipa-backup
        command backup the /etc/named/ipa-ext.conf file as well

        related : https://pagure.io/freeipa/issue/8079
        """
        # check of /etc/named/ipa-ext.conf exist
        assert self.master.transport.file_exists(paths.NAMED_CUSTOM_CONF)

        # check if /etc/named.conf does not contain 'allow-recursion { any; };'
        string_to_check = 'allow-recursion { any; };'
        named_contents = self.master.get_file_contents(paths.NAMED_CONF,
                                                       encoding='utf-8')
        assert string_to_check not in named_contents

        # check if ipa-backup command backups the /etc/named/ipa-ext.conf
        result = self.master.run_command(['ipa-backup', '-v'])
        assert paths.NAMED_CUSTOM_CONF in result.stderr_text

    def test_install_kra(self):
        tasks.install_kra(self.master, first_instance=True)

    def test_installer_wizard_prompts_for_DNS(self, server_cleanup):
        """
        Installer wizard should prompt for DNS even if --setup-dns is not
        provided as an argument.

        https://pagure.io/freeipa/issue/2575
        """
        cmd = ['ipa-server-install']
        netbios = create_netbios_name(self.master)
        with self.master.spawn_expect(cmd) as e:
            e.expect_exact('Do you want to configure integrated '
                           'DNS (BIND)? [no]: ')
            e.sendline('yes')
            e.expect_exact('Server host name [{}]: '.format(
                self.master.hostname))
            e.sendline(self.master.hostname)
            e.expect_exact('Please confirm the domain name [{}]: '.format(
                self.master.domain.name))
            e.sendline(self.master.domain.name)
            e.expect_exact('Please provide a realm name [{}]: '.format(
                self.master.domain.realm
            ))
            e.sendline(self.master.domain.realm)

            e.expect_exact('Directory Manager password: ')
            e.sendline(self.master.config.dirman_password)
            e.expect_exact('Password (confirm): ')
            e.sendline(self.master.config.dirman_password)

            e.expect_exact('IPA admin password: ')
            e.sendline(self.master.config.admin_password)
            e.expect_exact('Password (confirm): ')
            e.sendline(self.master.config.admin_password)

            e.expect_exact(
                'Do you want to configure DNS forwarders? [yes]: ')
            e.sendline('no')  # irrelevant for this test
            e.expect_exact('Do you want to search for missing reverse '
                           'zones? [yes]: ')
            e.sendline('no')  # irrelevant for this test
            e.expect_exact('NetBIOS domain name [{}]: '.format(netbios))
            e.sendline(netbios)
            e.expect_exact('Do you want to configure chrony with NTP '
                           'server or pool address? [no]: ')
            e.sendline('no')  # irrelevant for this test
            e.expect_exact('Continue to configure the system with these '
                           'values? [no]: ')
            e.sendline('yes')
            e.expect_exit(ignore_remaining_output=True, timeout=720)

        tasks.kinit_admin(self.master)
        result = self.master.run_command(
            ['ipa', 'dnszone-show', self.master.domain.name]
        ).stdout_text

        assert 'Active zone: TRUE' in result


class TestInstallMasterDNSRepeatedly(IntegrationTest):
    """ Test that a repeated installation of the primary with DNS enabled
    will lead to a already installed message and not in "DNS zone X
    already exists in DNS" in check_zone_overlap.
    The error is only occuring if domain is set explicitly in the command
    line installer as check_zone_overlap is used in the domain_name
    validator.
    """

    num_replicas = 0

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_dns=True)

    def test_install_master_releatedly(self):
        cmd = tasks.install_master(self.master,
                                   setup_dns=True,
                                   raiseonerr=False)
        exp_str = ("already exists in DNS")
        assert (exp_str not in cmd.stderr_text and cmd.returncode != 2)


class TestInstallMasterReservedIPasForwarder(IntegrationTest):
    """Test to check if IANA reserved IP doesn't accepted as DNS forwarder

    IANA reserved IP address can not be used as a forwarder.
    This test checks if ipa server installation throws an error when
    0.0.0.0 is specified as forwarder IP address.

    related ticket: https://pagure.io/freeipa/issue/6894
    """

    def test_reserved_ip_as_forwarder(self):
        args = [
            'ipa-server-install',
            '-n', self.master.domain.name,
            '-r', self.master.domain.realm,
            '-p', self.master.config.dirman_password,
            '-a', self.master.config.admin_password,
            '--setup-dns',
            '--forwarder', '0.0.0.0',
            '--auto-reverse']
        cmd = self.master.run_command(args, raiseonerr=False)
        assert cmd.returncode == 2
        exp_str = ("error: option --forwarder: invalid IP address 0.0.0.0: "
                   "cannot use IANA reserved IP address 0.0.0.0")
        assert exp_str in cmd.stderr_text

        server_install_options = (
                "yes\n"
                "{hostname}\n"
                "{dmname}\n\n"
                "{dm_pass}\n{dm_pass}"
                "\n{admin_pass}\n{admin_pass}\n"
                "yes\nyes\n0.0.0.0\n".format(
                    dm_pass=self.master.config.dirman_password,
                    admin_pass=self.master.config.admin_password,
                    dmname=self.master.domain.name,
                    hostname=self.master.hostname))

        cmd = self.master.run_command(['ipa-server-install'],
                                      stdin_text=server_install_options,
                                      raiseonerr=False)
        exp_str = ("Invalid IP Address 0.0.0.0: cannot use IANA reserved "
                   "IP address 0.0.0.0")
        assert exp_str in cmd.stdout_text


class TestKRAinstallAfterCertRenew(IntegrationTest):
    """ Test KRA installtion after ca agent cert renewal

    KRA installation was failing after ca-agent cert gets renewed.
    This test checks if the symptoms no longer exist.

    related ticket: https://pagure.io/freeipa/issue/7288
    """

    def test_KRA_install_after_cert_renew(self):

        tasks.install_master(self.master)

        # get ca-agent cert and load as pem
        dm_pass = self.master.config.dirman_password
        admin_pass = self.master.config.admin_password
        args = [paths.OPENSSL, "pkcs12", "-in",
                paths.DOGTAG_ADMIN_P12, "-nodes",
                "-passin", "pass:{}".format(dm_pass)]
        cmd = self.master.run_command(args)

        certs = x509.load_certificate_list(cmd.stdout_text.encode('utf-8'))

        # get expiry date of agent cert
        cert_expiry = certs[0].not_valid_after

        # move date to grace period so that certs get renewed
        self.master.run_command(['systemctl', 'stop', 'chronyd'])
        grace_date = cert_expiry - timedelta(days=10)
        grace_date = datetime.strftime(grace_date, "%Y-%m-%d %H:%M:%S")
        self.master.run_command(['date', '-s', grace_date])

        # get the count of certs track by certmonger
        cmd = self.master.run_command(['getcert', 'list'])
        cert_count = cmd.stdout_text.count('Request ID')
        timeout = 600
        count = 0
        start = time.time()
        # wait sometime for cert renewal
        while time.time() - start < timeout:
            cmd = self.master.run_command(['getcert', 'list'])
            count = cmd.stdout_text.count('status: MONITORING')
            if count == cert_count:
                break
            time.sleep(100)
        else:
            # timeout
            raise AssertionError('TimeOut: Failed to renew all the certs')

        # move date after 3 days of actual expiry
        cert_expiry = cert_expiry + timedelta(days=3)
        cert_expiry = datetime.strftime(cert_expiry, "%Y-%m-%d %H:%M:%S")
        self.master.run_command(['date', '-s', cert_expiry])

        passwd = "{passwd}\n{passwd}\n{passwd}".format(passwd=admin_pass)
        self.master.run_command(['kinit', 'admin'], stdin_text=passwd)
        cmd = self.master.run_command(['ipa-kra-install', '-p', dm_pass, '-U'])
        self.master.run_command(['systemctl', 'start', 'chronyd'])


class TestKRAinstallOnReplicaWithCAHost(IntegrationTest):
    """ Test that KRA install on replica with ca_host overriden fails

    KRA install on a replica should fail
    if the ca_host line in /etc/ipa/default.conf is present

    Related: https://pagure.io/freeipa/issue/8245
    """

    num_replicas = 1

    def test_kra_install_on_replica_with_ca_host_overriden(self):
        tasks.install_master(self.master)
        tasks.install_replica(self.master, self.replicas[0])

        content = self.replicas[0].get_file_contents(paths.IPA_DEFAULT_CONF,
                                                     encoding='utf-8')
        ca_host_line = "ca_host = %s" % self.master.hostname
        new_content = content + '\n' + ca_host_line
        self.replicas[0].put_file_contents(paths.IPA_DEFAULT_CONF,
                                           new_content)

        self.master.run_command(['firewall-cmd', '--add-port=8443/tcp'])

        result = tasks.install_kra(self.replicas[0], raiseonerr=False)

        err_str = "KRA can not be installed when 'ca_host' is overriden in IPA"
        "configuration file."
        assert result.returncode == 1
        assert err_str in result.stderr_text


class TestMaskInstall(IntegrationTest):
    """ Test master and replica installation with wrong mask

    This test checks that master/replica installation fails (expectedly) if
    mask > 022.

    related ticket: https://pagure.io/freeipa/issue/7193
    """
    topology = 'star'

    @classmethod
    def install(cls, mh):
        super(TestMaskInstall, cls).install(mh)
        cls.bashrc_file = cls.master.get_file_contents('/root/.bashrc')

    def test_install_master(self):
        self.master.run_command('echo "umask 0027" >> /root/.bashrc')
        result = self.master.run_command(['umask'])
        assert '0027' in result.stdout_text

        cmd = tasks.install_master(
            self.master, setup_dns=False, raiseonerr=False
        )
        exp_str = ("Unexpected system mask")
        assert (exp_str in cmd.stderr_text and cmd.returncode != 0)

    def test_install_replica(self):
        result = self.master.run_command(['umask'])
        assert '0027' in result.stdout_text

        cmd = self.master.run_command([
            'ipa-replica-install', '-w', self.master.config.admin_password,
            '-n', self.master.domain.name, '-r', self.master.domain.realm,
            '--server', 'dummy_master.%s' % self.master.domain.name,
            '-U'], raiseonerr=False
        )
        exp_str = ("Unexpected system mask")
        assert (exp_str in cmd.stderr_text and cmd.returncode != 0)

    def test_files_ownership_and_permission_teardown(self):
        """ Method to restore the default bashrc contents"""
        if self.bashrc_file is not None:
            self.master.put_file_contents('/root/.bashrc', self.bashrc_file)


class TestInstallMasterReplica(IntegrationTest):
    """https://pagure.io/freeipa/issue/7929
    Problem:
    If a replica installation fails before all the services
    have been enabled then
    it could leave things in a bad state.

    ipa-replica-manage del --cleanup --force
    invalid 'PKINIT enabled server': all masters must have
    IPA master role enabled

    Root cause was that configuredServices were being
    considered when determining what masters provide
    what services, so a partially installed master
    could cause operations to fail on other masters,
    to the point where a broken master couldn't be removed.
    """
    num_replicas = 1
    topology = 'star'

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_kra=True)
        # do not install KRA on replica, it is part of test
        tasks.install_replica(cls.master, cls.replicas[0], setup_kra=False)

    def test_replicamanage_del(self):
        """Test Steps:
        1. Setup server
        2. Setup replica
        3. modify the replica entry on Master:
           ldapmodify -D cn="Directory Manager"-w <passwd>
           dn: cn=KDC,cn=<replicaFQDN>,cn=masters,cn=ipa,cn=etc,<baseDN>
           changetype: modify
           delete: ipaconfigstring
           ipaconfigstring: enabledService

           dn: cn=KDC,cn=<replicaFQDN>,cn=masters,cn=ipa,cn=etc,<baseDN>
           add: ipaconfigstring
           ipaconfigstring: configuredService
        4. On master,
           run ipa-replica-manage del <replicaFQDN> --cleanup --force
        """
        # https://pagure.io/freeipa/issue/7929
        # modify the replica entry on Master
        cmd_output = None
        dn_entry = 'dn: cn=KDC,cn=%s,cn=masters,cn=ipa,' \
                   'cn=etc,%s' % \
                   (self.replicas[0].hostname,
                    ipautil.realm_to_suffix(
                        self.replicas[0].domain.realm).ldap_text())
        entry_ldif = textwrap.dedent("""
            {dn}
            changetype: modify
            delete: ipaconfigstring
            ipaconfigstring: enabledService

            {dn}
            add: ipaconfigstring
            ipaconfigstring: configuredService
        """).format(dn=dn_entry)
        cmd_output = tasks.ldapmodify_dm(self.master, entry_ldif)
        assert 'modifying entry' in cmd_output.stdout_text

        cmd_output = self.master.run_command([
            'ipa-replica-manage', 'del',
            self.replicas[0].hostname, '--cleanup', '--force'
        ])

        assert_text = 'Deleted IPA server "%s"' % self.replicas[0].hostname
        assert assert_text in cmd_output.stdout_text


class TestInstallReplicaAgainstSpecificServer(IntegrationTest):
    """Installation of replica against a specific server

    Test to check replica install against specific server. It uses master and
    replica1 without CA and having custodia service stopped. Then try to
    install replica2 from replica1 and expect it to get fail as specified
    server is not providing all the services.

    related ticket: https://pagure.io/freeipa/issue/7566
    """

    num_replicas = 2

    @classmethod
    def install(cls, mh):
        tasks.install_master(cls.master, setup_kra=True)

        # install replica1 without CA
        cmd = tasks.install_replica(cls.master, cls.replicas[0],
                                    setup_ca=False, setup_dns=True,
                                    promote=False)

        # check for warning that CA is not installed on server
        warn = 'WARNING: The CA service is only installed on one server'
        assert warn in cmd.stderr_text

    def test_replica_install_against_server_without_ca(self):
        """Replica install will fail complaining about CA role
        and exit code 4"""

        # stop custodia service on replica1
        self.replicas[0].run_command('systemctl stop ipa-custodia.service')

        # check if custodia service is stopped
        cmd = self.replicas[0].run_command('ipactl status', raiseonerr=False)
        assert 'ipa-custodia Service: STOPPED' in cmd.stdout_text

        try:
            # install replica2 against replica1, as CA is not installed on
            # replica1, installation on replica2 should fail
            cmd = tasks.install_replica(self.replicas[0], self.replicas[1],
                                        promote=False, raiseonerr=False)
            assert cmd.returncode == 4
            error = "please provide a server with the CA role"
            assert error in cmd.stderr_text

        finally:
            tasks.uninstall_master(self.replicas[1],
                                   ignore_topology_disconnect=True,
                                   ignore_last_of_role=True)

    def test_replica_install_against_server_without_kra(self):
        """Replica install will fail complaining about KRA role
        and exit code 4"""

        # install ca on replica1
        tasks.install_ca(self.replicas[0])
        try:
            # install replica2 against replica1, as KRA is not installed on
            # replica1(CA installed), installation should fail on replica2
            cmd = tasks.install_replica(self.replicas[0], self.replicas[1],
                                        promote=False, setup_kra=True,
                                        raiseonerr=False)
            assert cmd.returncode == 4
            error = "please provide a server with the KRA role"
            assert error in cmd.stderr_text

        finally:
            tasks.uninstall_master(self.replicas[1],
                                   ignore_topology_disconnect=True,
                                   ignore_last_of_role=True)

    def test_replica_install_against_server(self):
        """Replica install should succeed if specified server provide all
        the services"""

        tasks.install_replica(self.master, self.replicas[1],
                              setup_dns=True, promote=False)

        # check if replication agreement stablished between master
        # and replica2 only.
        cmd = self.replicas[1].run_command(['ipa-replica-manage', 'list',
                                            self.replicas[0].hostname])
        assert self.replicas[0].hostname not in cmd.stdout_text

        dirman_password = self.master.config.dirman_password
        cmd = self.replicas[1].run_command(['ipa-csreplica-manage', 'list',
                                            self.replicas[0].hostname],
                                           stdin_text=dirman_password)
        assert self.replicas[0].hostname not in cmd.stdout_text


class TestInstallWithoutSudo(IntegrationTest):

    num_clients = 1
    num_replicas = 1
    no_sudo_str = "The sudo binary does not seem to be present on this"
    sudo_version_str = "Sudo version"

    @classmethod
    def install(cls, mh):
        pass

    def test_sudo_removal(self):
        # ipa-client makes sudo depend on libsss_sudo.

        # --nodeps is mandatory because dogtag uses sudo at install
        # time until commit 49585867207922479644a03078c29548de02cd03
        # which is scheduled to land in 10.10.

        # This also means sudo+libsss_sudo cannot be uninstalled on
        # IPA servers with a CA.
        assert tasks.is_package_installed(self.clients[0], 'sudo')
        assert tasks.is_package_installed(self.clients[0], 'libsss_sudo')
        tasks.uninstall_packages(
            self.clients[0], ['sudo', 'libsss_sudo'], nodeps=True
        )

    def test_ipa_installation_without_sudo(self):
        # FixMe: When Dogtag 10.10 is out, test installation without sudo
        tasks.install_master(self.master, setup_dns=True)

    def test_replica_installation_without_sudo(self):
        # FixMe: When Dogtag 10.10 is out, test replica installation
        # without sudo and with CA
        tasks.uninstall_packages(
            self.replicas[0], ['sudo', 'libsss_sudo'], nodeps=True
        )
        # One-step install is needed.
        # With promote=True, two-step install is done and that only captures
        # the ipa-replica-install stdout/stderr, not ipa-client-install's.
        result = tasks.install_replica(
            self.master, self.replicas[0], promote=False,
            setup_dns=True, setup_ca=False
        )
        assert self.no_sudo_str in result.stderr_text

    def test_client_installation_without_sudo(self):
        result = tasks.install_client(self.master, self.clients[0])
        assert self.no_sudo_str in result.stderr_text

    def test_remove_sudo_on_ipa(self):
        tasks.uninstall_packages(
            self.master, ['sudo', 'libsss_sudo'], nodeps=True
        )
        self.master.run_command(
            ['ipactl', 'restart']
        )

    def test_install_sudo_on_client(self):
        """ Check that installing sudo pulls libsss_sudo in"""
        for pkg in ('sudo', 'libsss_sudo'):
            assert tasks.is_package_installed(self.clients[0], pkg) is False
        tasks.uninstall_client(self.clients[0])
        tasks.install_packages(self.clients[0], ['sudo'])
        for pkg in ('sudo', 'libsss_sudo'):
            assert tasks.is_package_installed(self.clients[0], pkg)
        result = tasks.install_client(self.master, self.clients[0])
        assert self.no_sudo_str not in result.stderr_text
        assert self.sudo_version_str not in result.stdout_text


class TestInstallWithoutNamed(IntegrationTest):
    num_replicas = 1

    @classmethod
    def remove_named(cls, host):
        # remove the bind package and make sure the named user does not exist.
        # https://pagure.io/freeipa/issue/8936
        result = host.run_command(['id', 'named'], raiseonerr=False)
        if result.returncode == 0:
            tasks.uninstall_packages(host, ['bind'])
            host.run_command(['userdel', constants.NAMED_USER])
        assert host.run_command(
            ['id', 'named'], raiseonerr=False
        ).returncode == 1

    @classmethod
    def install(cls, mh):
        for tgt in (cls.master, cls.replicas[0]):
            cls.remove_named(tgt)
        tasks.install_master(cls.master, setup_dns=False)

    def test_replica0_install(self):
        tasks.install_replica(
            self.master, self.replicas[0], setup_ca=False, setup_dns=False
        )



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

from __future__ import absolute_import

import re
from contextlib import contextmanager

import pytest

from ipatests.pytest_ipa.integration import tasks
from ipatests.pytest_ipa.integration.firewall import Firewall
from ipatests.test_integration.base import IntegrationTest
from ipaplatform.paths import paths


class TestHttpKdcProxy(IntegrationTest):
    topology = "line"
    num_clients = 1
    num_ad_domains = 1

    @classmethod
    def install(cls, mh):
        super().install(mh)

        cls.client = cls.clients[0]
        cls.ad = cls.ads[0]

        tasks.kinit_admin(cls.master)
        cls.master.run_command(['ipa', 'pwpolicy-mod', '--minlife=0'])

        tasks.install_adtrust(cls.master)
        tasks.configure_dns_for_trust(cls.master, cls.ad)
        tasks.establish_trust_with_ad(cls.master, cls.ad.domain.name)

    @classmethod
    def uninstall(cls, mh):
        tasks.remove_trust_info_from_ad(
            cls.master, cls.ad.domain.name, cls.ad.hostname)
        super().uninstall(mh)

    @pytest.fixture(autouse=True, scope='function')
    def cleanup_credentials(self):
        tasks.kdestroy_all(self.client)
        tasks.clear_sssd_cache(self.client)

    @pytest.fixture(scope='class')
    def users(self, mh):
        master = mh.master
        ad = mh.ads[0]
        users = {
            'ipa': {
                'name': 'ipa_test_user',
                'password': 'SecretIpaTestUser',
                'domain': mh.master.domain,
                'test_service': 'HTTP/{}@{}'
                .format(master.hostname, master.domain.realm),
            },
            'ad': {
                'name': 'testuser@{}'.format(ad.domain.realm),
                'password': 'Secret123',
                'domain': ad.domain,
                'test_service': 'HTTP/{}@{}'
                .format(ad.hostname, ad.domain.realm),
            }
        }
        tasks.kinit_admin(mh.master)
        tasks.create_active_user(
            mh.master, users['ipa']['name'], users['ipa']['password'])
        yield users
        tasks.kinit_admin(mh.master)
        mh.master.run_command(['ipa', 'user-del', users['ipa']['name']])

    @pytest.fixture()
    def restrict_network_for_client(self, mh):
        fw_rules_allow = [
            ['OUTPUT', '-p', 'udp', '--dport', '53', '-j', 'ACCEPT'],
            ['OUTPUT', '-p', 'tcp', '--dport', '80', '-j', 'ACCEPT'],
            ['OUTPUT', '-p', 'tcp', '--dport', '443', '-j', 'ACCEPT'],
            ['OUTPUT', '-p', 'tcp', '--sport', '22', '-j', 'ACCEPT']]
        fw = Firewall(self.client)
        fw.prepend_passthrough_rules(fw_rules_allow)
        fw.passthrough_rule(['-P', 'OUTPUT', 'DROP'])
        yield
        fw.passthrough_rule(['-P', 'OUTPUT', 'ACCEPT'])
        fw.remove_passthrough_rules(fw_rules_allow)

    @pytest.fixture()
    def client_use_kdcproxy(self, mh):
        """Configure client for using kdcproxy for IPA and AD domains."""
        def replace_regexp_once(pattern, repl, string):
            res, n = re.subn(pattern, repl, string)
            assert n == 1
            return res

        krb5conf_backup = tasks.FileBackup(self.client, paths.KRB5_CONF)
        krb5conf = self.client.get_file_contents(
            paths.KRB5_CONF, encoding='utf-8')
        kdc_url = 'https://{}/KdcProxy'.format(self.master.hostname)

        # configure kdc proxy for IPA realm
        krb5conf = replace_regexp_once(
            r' kdc = .+', ' kdc = {}'.format(kdc_url), krb5conf)
        krb5conf = replace_regexp_once(
            r'kpasswd_server = .+', 'kpasswd_server = {}'.format(kdc_url),
            krb5conf)

        # configure kdc proxy for Windows AD realm
        ad_realm_config = '''
        {realm} = {{
            kdc = {kdc_url}
            kpasswd_server = {kdc_url}
        }}
        '''.format(realm=self.ad.domain.realm, kdc_url=kdc_url)
        krb5conf = replace_regexp_once(
            r'\[realms\]',
            '[realms]' + ad_realm_config,
            krb5conf
        )

        self.client.put_file_contents(paths.KRB5_CONF, krb5conf)
        self.client.run_command(['systemctl', 'restart', 'sssd.service'])
        yield
        krb5conf_backup.restore()
        self.client.run_command(['systemctl', 'restart', 'sssd.service'])

    @contextmanager
    def configure_kdc_proxy_for_ad_trust(self, use_tcp):
        backup = tasks.FileBackup(self.master, paths.KDCPROXY_CONFIG)
        with tasks.remote_ini_file(self.master, paths.KDCPROXY_CONFIG) as conf:
            conf.set('global', 'use_dns', 'true')
            conf.set('global', 'configs', 'mit')
            if use_tcp:
                conf.add_section(self.ad.domain.realm)
                conf.set(self.ad.domain.realm, 'kerberos',
                         'kerberos+tcp://{}:88'.format(self.ad.hostname))
                conf.set(self.ad.domain.realm, 'kpasswd',
                         'kpasswd+tcp://{}:464'.format(self.ad.hostname))
        try:
            self.master.run_command(['ipactl', 'restart'])
            yield
        finally:
            backup.restore()
            self.master.run_command(['ipactl', 'restart'])

    def check_kerberos_requests(self, user, skip_kpasswd_check=False):
        # KDC AS request
        tasks.kinit_as_user(self.client, user['name'], user['password'])

        # KDC TGS requests
        self.client.run_command(['kvno', user['test_service']])

        # KDC AS requests and kpasswd requests

        # Changing password on Windows AD can not be done now because
        # of default password policy mandating that minimal password lifetime
        # is one day.
        # Once we switch to dynamically creating test users in Windows AD
        # and create an utility for modifying Group Policy Objects then
        # we should update test setup and remove this condition.
        def set_password(old_pass, new_pass):
            with self.client.spawn_expect(['kpasswd', user['name']]) as e:
                e.expect('Password for .+:')
                e.sendline(old_pass)
                e.expect_exact('Enter new password:')
                e.sendline(new_pass)
                e.expect_exact('Enter it again:')
                e.sendline(new_pass)
                e.expect_exit(ignore_remaining_output=True)

        if not skip_kpasswd_check:
            test_password = 'Secret123456'
            set_password(user['password'], test_password)
            # Restore password:
            set_password(test_password, user['password'])

    @pytest.mark.parametrize('user_origin', ['ipa', 'ad'])
    def test_user_login_on_client_without_firewall(self, users, user_origin):
        """Basic check for test setup."""
        self.check_kerberos_requests(users[user_origin],
                                     skip_kpasswd_check=user_origin == 'ad')

    @pytest.mark.usefixtures('restrict_network_for_client')
    @pytest.mark.parametrize('user_origin', ['ipa', 'ad'])
    def test_access_blocked_on_client_without_kdcproxy(
            self, users, user_origin):
        """Check for test firewall setup."""
        user = users[user_origin]
        result = tasks.kinit_as_user(
            self.client, user['name'], user['password'], raiseonerr=False)
        expected_errors = [
            ("Cannot contact any KDC for realm '{}' while getting initial "
             "credentials".format(user['domain'].realm)),
            ('Cannot find KDC for realm "{}" while getting initial '
             "credentials".format(user['domain'].realm)),
        ]
        assert (result.returncode == 1
                and any(s in result.stderr_text for s in expected_errors))

    @pytest.mark.usefixtures('restrict_network_for_client',
                             'client_use_kdcproxy')
    def test_ipa_user_login_on_client_with_kdcproxy(self, users):
        self.check_kerberos_requests(users['ipa'])

    @pytest.mark.usefixtures('restrict_network_for_client',
                             'client_use_kdcproxy')
    @pytest.mark.parametrize('use_tcp', [True, False])
    def test_ad_user_login_on_client_with_kdcproxy(self, users, use_tcp):
        with self.configure_kdc_proxy_for_ad_trust(use_tcp):
            self.check_kerberos_requests(users['ad'], skip_kpasswd_check=True)

    @pytest.fixture()
    def windows_small_mtu_size(self, mh):
        new_mtu = 70

        def get_iface_name():
            result = self.ad.run_command([
                'powershell', '-c',
                '(Get-NetIPAddress -IPAddress {}).InterfaceAlias'.format(
                    self.ad.ip)])
            return result.stdout_text.strip()

        def get_mtu(iface_name):
            result = self.ad.run_command([
                'netsh', 'interface', 'ipv4', 'show', 'subinterface',
                iface_name])
            mtu = result.stdout_text.strip().splitlines()[-1].split()[0]
            return int(mtu)

        def set_mtu(iface_name, mtu):
            self.ad.run_command([
                'netsh', 'interface', 'ipv4', 'set', 'subinterface',
                iface_name, 'mtu={}'.format(mtu)])

        iface_name = get_iface_name()
        original_mtu = get_mtu(iface_name)
        set_mtu(iface_name, new_mtu)
        # `netsh` does not report failures with return code so we check
        # it was successful by inspecting the actual value of MTU
        assert get_mtu(iface_name) == new_mtu
        yield
        set_mtu(iface_name, original_mtu)
        assert get_mtu(iface_name) == original_mtu

    @pytest.mark.usefixtures('restrict_network_for_client',
                             'client_use_kdcproxy',
                             'windows_small_mtu_size')
    def test_kdcproxy_handles_small_packets_from_ad(self, users):
        """Check that kdcproxy handles AD response split to several TCP packets

        This is a regression test for the bug in python-kdcproxy:
        https://github.com/latchset/kdcproxy/pull/44
          When the reply from AD is split into several TCP packets the kdc
          proxy software cannot handle it and returns a false error message
          indicating it cannot contact the KDC server.
        """
        with self.configure_kdc_proxy_for_ad_trust(use_tcp=True):
            self.check_kerberos_requests(users['ad'], skip_kpasswd_check=True)



#
# Copyright (C) 2020  FreeIPA Contributors see COPYING for license
#
"""Misc test for 'ipa' CLI regressions
"""
from __future__ import absolute_import

import pytest

from ipatests.test_integration.base import IntegrationTest

from ipatests.pytest_ipa.integration import tasks

USER = 'tuser'
PASSWORD = 'Secret123'
POLICY = 'test'


class TestPWPolicy(IntegrationTest):
    """
    Test password policy in action.
    """

    topology = 'line'

    @classmethod
    def install(cls, mh):
        super(TestPWPolicy, cls).install(mh)

        tasks.kinit_admin(cls.master)
        cls.master.run_command(['ipa', 'user-add', USER,
                                '--first', 'Test',
                                '--last', 'User'])
        cls.master.run_command(['ipa', 'group-add', POLICY])
        cls.master.run_command(['ipa', 'group-add-member', POLICY,
                                '--users', USER])
        cls.master.run_command(['ipa', 'pwpolicy-add', POLICY,
                                '--priority', '1'])
        cls.master.run_command(['ipa', 'passwd', USER],
                               stdin_text='{password}\n{password}\n'.format(
                               password=PASSWORD
                               ))

    def kinit_as_user(self, host, old_password, new_password, user=USER,
                      raiseonerr=True):
        """kinit to an account with an expired password"""
        return host.run_command(
            ['kinit', user],
            raiseonerr=raiseonerr,
            stdin_text='{old}\n{new}\n{new}\n'.format(
                old=old_password, new=new_password
            ),
        )

    def reset_password(self, host, user=USER, password=PASSWORD):
        tasks.kinit_admin(host)
        host.run_command(
            ['ipa', 'passwd', user],
            stdin_text='{password}\n{password}\n'.format(password=password),
        )

    def set_pwpolicy(self, minlength=None, maxrepeat=None, maxsequence=None,
                     dictcheck=None, usercheck=None, minclasses=None):
        tasks.kinit_admin(self.master)
        args = ["ipa", "pwpolicy-mod", POLICY]
        if minlength is not None:
            args.append("--minlength={}".format(minlength))
        if maxrepeat is not None:
            args.append("--maxrepeat={}".format(maxrepeat))
        if maxsequence is not None:
            args.append("--maxsequence={}".format(maxsequence))
        if dictcheck is not None:
            args.append("--dictcheck={}".format(dictcheck))
        if usercheck is not None:
            args.append("--usercheck={}".format(usercheck))
        if minclasses is not None:
            args.append("--minclasses={}".format(minclasses))
        self.master.run_command(args)

        self.reset_password(self.master)

    def clean_pwpolicy(self):
        """Set all policy values we care about to zero/false"""
        self.master.run_command(
            ["ipa", "pwpolicy-mod", POLICY,
             "--maxrepeat", "0",
             "--maxsequence", "0",
             "--usercheck", "false",
             "--dictcheck" ,"false",
             "--minlife", "0",
             "--minlength", "0",
             "--minclasses", "0",],
        )

    @pytest.fixture
    def reset_pwpolicy(self):
        """Fixture to ensure policy is reset between tests"""
        yield
        tasks.kinit_admin(self.master)
        self.clean_pwpolicy()

    def test_maxrepeat(self, reset_pwpolicy):
        self.set_pwpolicy(maxrepeat=2)
        # good passwords
        for password in ('Secret123', 'Password'):
            self.reset_password(self.master)
            self.kinit_as_user(self.master, PASSWORD, password)
            self.reset_password(self.master)
            tasks.ldappasswd_user_change(USER, PASSWORD, password, self.master)

        self.reset_password(self.master)

        # bad passwords
        for password in ('Secret1111', 'passsword'):
            result = self.kinit_as_user(self.master, PASSWORD, password,
                                        raiseonerr=False)
            assert result.returncode == 1
            result = tasks.ldappasswd_user_change(USER, PASSWORD, password,
                                                  self.master,
                                                  raiseonerr=False)
            assert result.returncode == 1
            assert 'Password has too many consecutive characters' in \
                result.stdout_text

    def test_maxsequence(self, reset_pwpolicy):
        self.set_pwpolicy(maxsequence=3)
        # good passwords
        for password in ('Password123', 'Passwordabc'):
            self.reset_password(self.master)
            self.kinit_as_user(self.master, PASSWORD, password)
            self.reset_password(self.master)
            tasks.ldappasswd_user_change(USER, PASSWORD, password, self.master)

        self.reset_password(self.master)

        # bad passwords
        for password in ('Password1234', 'Passwordabcde'):
            result = self.kinit_as_user(self.master, PASSWORD, password,
                                        raiseonerr=False)
            assert result.returncode == 1
            result = tasks.ldappasswd_user_change(USER, PASSWORD, password,
                                                  self.master,
                                                  raiseonerr=False)
            assert result.returncode == 1
            assert 'Password contains a monotonic sequence' in \
                result.stdout_text

    def test_usercheck(self, reset_pwpolicy):
        self.set_pwpolicy(usercheck=True)
        for password in ('tuserpass', 'passoftuser'):
            result = self.kinit_as_user(self.master, PASSWORD, password,
                                        raiseonerr=False)
            assert result.returncode == 1
            result = tasks.ldappasswd_user_change(USER, PASSWORD, password,
                                                  self.master,
                                                  raiseonerr=False)
            assert result.returncode == 1
            assert 'Password contains username' in \
                result.stdout_text

        # test with valid password
        self.kinit_as_user(self.master, PASSWORD, 'bamOncyftAv0')

    def test_dictcheck(self, reset_pwpolicy):
        self.set_pwpolicy(dictcheck=True)
        for password in ('password', 'bookends', 'BaLtim0re'):
            result = self.kinit_as_user(self.master, PASSWORD, password,
                                        raiseonerr=False)
            assert result.returncode == 1
            result = tasks.ldappasswd_user_change(USER, PASSWORD, password,
                                                  self.master,
                                                  raiseonerr=False)
            assert result.returncode == 1
            assert 'Password is based on a dictionary word' in \
                result.stdout_text

        # test with valid password
        self.kinit_as_user(self.master, PASSWORD, 'bamOncyftAv0')

    def test_minclasses(self, reset_pwpolicy):
        self.set_pwpolicy(minclasses=2)
        for password in ('password', 'bookends'):
            result = self.kinit_as_user(self.master, PASSWORD, password,
                                        raiseonerr=False)
            assert result.returncode == 1
            assert 'Password does not contain enough character' in \
                result.stdout_text
            result = tasks.ldappasswd_user_change(USER, PASSWORD, password,
                                                  self.master,
                                                  raiseonerr=False)
            assert result.returncode == 1
            assert 'Password is too simple' in \
                result.stdout_text

        # test with valid password
        for valid in ('Password', 'password1', 'password!'):
            self.kinit_as_user(self.master, PASSWORD, valid)
            self.reset_password(self.master)

        self.set_pwpolicy(minclasses=3)
        for password in ('password1', 'Bookends'):
            result = self.kinit_as_user(self.master, PASSWORD, password,
                                        raiseonerr=False)
            assert result.returncode == 1
            assert 'Password does not contain enough character' in \
                result.stdout_text
            result = tasks.ldappasswd_user_change(USER, PASSWORD, password,
                                                  self.master,
                                                  raiseonerr=False)
            assert result.returncode == 1
            assert 'Password is too simple' in \
                result.stdout_text

        # test with valid password
        for valid in ('Passw0rd', 'password1!', 'Password!'):
            self.kinit_as_user(self.master, PASSWORD, valid)
            self.reset_password(self.master)

    def test_minlength_mod(self, reset_pwpolicy):
        """Test that the pwpolicy minlength overrides our policy
        """

        # With a minlength of 4 all settings of pwq should fail
        self.master.run_command(
            ["ipa", "pwpolicy-mod", POLICY,
             "--minlength", "4",]
        )
        for values in (('--maxrepeat', '4'),
                       ('--maxsequence', '4'),
                       ('--dictcheck', 'true'),
                       ('--usercheck', 'true')):
            args = ["ipa", "pwpolicy-mod", POLICY]
            args.extend(values)
            result = self.master.run_command(args, raiseonerr=False)
            assert result.returncode != 0
            assert 'minlength' in result.stderr_text

        # With any pwq value set, setting minlife < 6 should fail
        for values in (('--maxrepeat', '4'),
                       ('--maxsequence', '4'),
                       ('--dictcheck', 'true'),
                       ('--usercheck', 'true')):
            self.clean_pwpolicy()
            args = ["ipa", "pwpolicy-mod", POLICY]
            args.extend(values)
            self.master.run_command(args)
            result = self.master.run_command(
                ["ipa", "pwpolicy-mod", POLICY,
                 "--minlength", "4",], raiseonerr=False
            )
            assert result.returncode != 0
            assert 'minlength' in result.stderr_text

    def test_minlength_add(self):
        """Test that adding a new policy with minlength is caught.
        """
        result = self.master.run_command(
            ["ipa", "pwpolicy-add", "test_add",
             "--maxrepeat", "4", "--minlength", "4", "--priority", "2"],
            raiseonerr=False
        )
        assert result.returncode != 0
        assert 'minlength' in result.stderr_text
